import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # or any {'0', '1', '2'}
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.WARN)  # or any {DEBUG, INFO, WARN, ERROR, FATAL}
from tensorflow.keras.layers import Conv1D, MaxPooling1D, Flatten, Dense, Dropout, Input
from tensorflow.keras.models import Model




class M1CNN(Model):
    def __init__(self, payload_size=784, n_classes=2) -> None:
        super(M1CNN, self).__init__()
        input_layer = Input(shape=(payload_size,1))
        self.model = Model(
            name='M1CNN',
            inputs=input_layer,
            outputs=stack([
                input_layer, # first layer
                Conv1D(32, 25, strides=1, padding="same", activation='relu'),
                MaxPooling1D(3, strides=3, padding="same",),
                Conv1D(64, 25, strides=1, padding="same", activation='relu'),
                MaxPooling1D(3, strides=3, padding="same",),
                Flatten(),
                Dense(1024, activation='relu'),
                Dropout(0.2),
                Dense(n_classes, activation='softmax'),
            ])
        )
        
        '''
        optimizer='adam',
        loss=tf.keras.losses.CategoricalCrossentropy(),
        '''
        
        
    def call(self, inputs, training=None):
        # See: https://www.tensorflow.org/guide/keras/custom_layers_and_models#the_model_class
        return self.model(inputs, training)



###################
# Model utilities #
###################

def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack


from hosts_processor import HostsProcessor
from stats.stats import IterableStats, StatsCollection
from nfstream import NFStreamer
import pandas as pd
import numpy as np


class FlowsProcessor:
    def __init__(self, streamer: NFStreamer) -> None:
        self.streamer = streamer

    def process(self):
        byte_count = 0
        flows = 0

        byte_frequency = np.zeros(256)
        prt_stats_list = list()

        tcp = 0
        udp = 0
        dns = 0
        
        for flow in self.streamer:
            flows += 1
            byte_count += np.sum(flow.udps.bidirectional_n_packets_byte_frequency)
            byte_frequency += flow.udps.bidirectional_n_packets_byte_frequency
            prts    = IterableStats(flow.udps.packet_relative_times)
            prrdts  = IterableStats(flow.udps.req_res_time_diff)
            prt_stats_list.append(prts)
            if flow.protocol == 6:
                tcp += flow.bidirectional_packets
            elif flow.protocol == 17:
                udp += flow.bidirectional_packets
                if flow.udps.bidirectional_dns_packets:
                    dns += flow.udps.bidirectional_dns_packets
            
        print(StatsCollection(prt_stats_list).average())

        print(flows)
        print('TCP: ', tcp, 'UDP:', udp, 'DNS:', dns)
        print(byte_frequency / byte_count) # frequency to distribution
        req_res =  list(map( 
                    lambda flows:
                        IterableStats(
                            sum(
                                map(
                                lambda flow: flow.udps.req_res_time_diff,
                                flows
                                ),
                                []
                            )
                        ),
                    nfc.sessions.values()
                    ))
        
        
        
        df = self.streamer.to_pandas(columns_to_anonymize=[])
        df.to_csv('out-flows.csv')
        # self._host_features(df)
        #print(df.head())
        #df.to_csv('out.csv')

    def _host_features(self, df):
        HostsProcessor().process(df)


    



from nfstream import NFStreamer, NFPlugin
import numpy as np
from scapy.all import IP, IPv6, raw

class NBytes(NFPlugin):
    '''
        Extracts the first n_bytes from the flow, the bytes are taken
        from the transport layer payload (L4). if the flow have less than n_bytes bytes,
        then the rest of the bytes are zero-valued.
    '''
    def __init__(self, n=784):
        self.n = n
    
    def on_init(self, packet, flow):
        flow.udps.n_bytes_value = self.n
        flow.udps.n_bytes = np.zeros(self.n)
        flow.udps.n_bytes_counted = 0
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        remaining_bytes = self.n - flow.udps.n_bytes_counted
        if remaining_bytes >= 0 and packet.protocol in [6, 17]:
            amount_to_copy = min(remaining_bytes, packet.payload_size)
            if amount_to_copy == 0:
                return
            max_index_to_copy = -packet.payload_size+amount_to_copy if -packet.payload_size+amount_to_copy != 0 else None
            '''
            print(type(remaining_bytes), type(amount_to_copy), type(flow.udps.n_bytes_counted))
            print(remaining_bytes, amount_to_copy, flow.udps.n_bytes_counted, packet.ip_packet, np.frombuffer(packet.ip_packet, dtype=np.uint8))
            print(-packet.payload_size, max_index_to_copy)
            print(flow.udps.n_bytes[flow.udps.n_bytes_counted:flow.udps.n_bytes_counted+amount_to_copy])
            print(packet.ip_packet[-packet.payload_size:max_index_to_copy])
            print(np.frombuffer(packet.ip_packet[-packet.payload_size:max_index_to_copy], dtype=np.uint8))
            '''
            flow.udps.n_bytes[flow.udps.n_bytes_counted:flow.udps.n_bytes_counted+amount_to_copy] = np.frombuffer(self.get_payload_as_binary(packet, flow.ip_version)[-packet.payload_size:max_index_to_copy], dtype=np.uint8)
            flow.udps.n_bytes_counted += amount_to_copy

    def on_expire(self, flow):
        flow.udps.n_bytes /= 255
        flow.udps.n_bytes = list(flow.udps.n_bytes)
        # optional: del flow.udps.n_bytes_counted

    def get_payload_as_binary(self, packet, ip_version):
            if ip_version == 4:
                scapy_packet = IP(packet.ip_packet)
            elif ip_version == 6:
                scapy_packet = IPv6(packet.ip_packet)

            return raw(scapy_packet.payload.payload)


from ._deepmal import DeepMALRawFlows
from ._custom_distiller import CustomDistiller
from ._m1cnn import M1CNN
from ._distiller import Distiller
from ._maldist import MalDist

__all__ = [
    'DeepMALRawFlows',
    'CustomDistiller',
    'M1CNN',
    'Distiller',
    'MalDist'
]


from nfstream import NFPlugin


class FlowTimeSlicer(NFPlugin):
    '''
        DOESNT WORK AS NFSTREAM DOESNT ALLOW TO EXPIRE
        A STREAM EXCLUDING THE CURRENT PACKET.
    '''
    def __init__(self, time: float, **kwargs):  # time in seconds
        super().__init__(**kwargs)
        self.time = time
        self.flow_last_slicing_times = dict()

    def on_init(self, packet, flow):
        flow_five_tuple_string = self._five_tuple_string_of(flow)
        if flow_five_tuple_string not in self.flow_last_slicing_times:
            self.flow_last_slicing_times[flow_five_tuple_string] = flow.bidirectional_first_seen_ms

    def on_update(self, packet, flow):
        flow_five_tuple_string = self._five_tuple_string_of(flow)
        flow_slice_start_time  = self.flow_last_slicing_times[flow_five_tuple_string]
        current_time = packet.time
        time_diff = current_time - flow_slice_start_time
        time_slice_ms = self.time * 1000

        if time_diff > time_slice_ms:
            flow.expiration_id = -1  # -1 value force expiration
            self.flow_last_slicing_times[flow_five_tuple_string] = current_time

    def _five_tuple_string_of(self, flow):
        return str.join('',
            sorted([
                flow.src_ip,
                str(flow.src_port),
                flow.dst_ip,
                str(flow.dst_port)]))



from nfstream import NFPlugin
from scapy.all import IP, DNS, IPv6
from stats.stats import IterableStats

class DNSCounter(NFPlugin):
    '''
    Description:
        DNS over UDP only, supports MDNS (Multicast DNS)
        
    Features (bidirectional and unidirectional per direction):
        dns_packets   
        dns_queries   
        dns_responses 
        dns_qd_count  
        dns_an_count
        dns_ns_count
        dns_ar_count
        dns_response_digit_count (+ mean)
        dns_response_alpha_count (+ mean)
        dns_response_hypens_count (+ mean)
        dns_response_dots_count (+ mean)
        dns_response_ip_count (+ mean)
        dns_response_ttls_s (seconds) (+ Stats)
        
    Prefixes:
        bidirectional_
        src2dst_
        dst2src_
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''
        # bidirectional
        flow.udps.bidirectional_dns_packets   = 0
        flow.udps.bidirectional_dns_queries   = 0
        flow.udps.bidirectional_dns_responses = 0
        flow.udps.bidirectional_dns_qd_count  = 0
        flow.udps.bidirectional_dns_an_count  = 0 # answer count
        flow.udps.bidirectional_dns_ns_count  = 0 # name server count
        flow.udps.bidirectional_dns_ar_count  = 0 # additional information count
        flow.udps.bidirectional_dns_response_digit_count  = 0 # count of digits in DNS response.
        flow.udps.bidirectional_dns_response_alpha_count  = 0 # count of alpha in DNS response.
        flow.udps.bidirectional_dns_response_hypens_count = 0 # count of hypens in DNS response.
        flow.udps.bidirectional_dns_response_dots_count   = 0 # count of dots in DNS response.
        flow.udps.bidirectional_dns_response_ip_count     = 0 # number ips in response.
        flow.udps.bidirectional_dns_response_ttls_s       = [] # record Time-To-Live values in seconds.
        flow.udps.bidirectional_mean_dns_response_digit_count  = 0 # mean relative to total number of responses
        flow.udps.bidirectional_mean_dns_response_alpha_count  = 0 
        flow.udps.bidirectional_mean_dns_response_hypens_count = 0 
        flow.udps.bidirectional_mean_dns_response_dots_count   = 0 
        flow.udps.bidirectional_mean_dns_response_ip_count     = 0 
        flow.udps.bidirectional_mean_dns_response_ttls_s             = None
        flow.udps.bidirectional_median_dns_response_ttls_s           = None
        flow.udps.bidirectional_stdev_dns_response_ttls_s            = None
        flow.udps.bidirectional_variance_dns_response_ttls_s         = None
        flow.udps.bidirectional_coeff_of_var_dns_response_ttls_s     = None
        flow.udps.bidirectional_skew_from_median_dns_response_ttls_s = None
        flow.udps.bidirectional_min_dns_response_ttls_s              = None
        flow.udps.bidirectional_max_dns_response_ttls_s              = None
        # src -> dst
        flow.udps.src2dst_dns_packets   = 0
        flow.udps.src2dst_dns_queries   = 0
        flow.udps.src2dst_dns_responses = 0
        flow.udps.src2dst_dns_qd_count  = 0
        flow.udps.src2dst_dns_an_count  = 0 # answer count
        flow.udps.src2dst_dns_ns_count  = 0 # name server count
        flow.udps.src2dst_dns_ar_count  = 0 # additional information count
        flow.udps.src2dst_dns_response_digit_count  = 0 # count of digits in DNS response.
        flow.udps.src2dst_dns_response_alpha_count  = 0 # count of alpha in DNS response.
        flow.udps.src2dst_dns_response_hypens_count = 0 # count of hypens in DNS response.
        flow.udps.src2dst_dns_response_dots_count   = 0 # count of dots in DNS response.
        flow.udps.src2dst_dns_response_ip_count     = 0 # number ips in response.
        flow.udps.src2dst_dns_response_ttls_s       = [] # record Time-To-Live values in seconds.
        flow.udps.src2dst_dns_response_ttls_s       = [] # record Time-To-Live values in seconds.
        flow.udps.src2dst_mean_dns_response_digit_count  = 0 # mean relative to total number of responses
        flow.udps.src2dst_mean_dns_response_alpha_count  = 0 
        flow.udps.src2dst_mean_dns_response_hypens_count = 0 
        flow.udps.src2dst_mean_dns_response_dots_count   = 0 
        flow.udps.src2dst_mean_dns_response_ip_count     = 0 
        flow.udps.src2dst_mean_dns_response_ttls_s             = None
        flow.udps.src2dst_median_dns_response_ttls_s           = None
        flow.udps.src2dst_stdev_dns_response_ttls_s            = None
        flow.udps.src2dst_variance_dns_response_ttls_s         = None
        flow.udps.src2dst_coeff_of_var_dns_response_ttls_s     = None
        flow.udps.src2dst_skew_from_median_dns_response_ttls_s = None
        flow.udps.src2dst_min_dns_response_ttls_s              = None
        flow.udps.src2dst_max_dns_response_ttls_s              = None
        # dst -> src
        flow.udps.dst2src_dns_packets   = 0
        flow.udps.dst2src_dns_queries   = 0
        flow.udps.dst2src_dns_responses = 0
        flow.udps.dst2src_dns_qd_count  = 0
        flow.udps.dst2src_dns_an_count  = 0 # answer count
        flow.udps.dst2src_dns_ns_count  = 0 # name server count
        flow.udps.dst2src_dns_ar_count  = 0 # additional information count
        flow.udps.dst2src_dns_response_digit_count  = 0 # count of digits in DNS response.
        flow.udps.dst2src_dns_response_alpha_count  = 0 # count of alpha in DNS response.
        flow.udps.dst2src_dns_response_hypens_count = 0 # count of hypens in DNS response.
        flow.udps.dst2src_dns_response_dots_count   = 0 # count of dots in DNS response.
        flow.udps.dst2src_dns_response_ip_count     = 0 # number ips in response.
        flow.udps.dst2src_dns_response_ttls_s       = [] # record Time-To-Live values in seconds.
        flow.udps.dst2src_mean_dns_response_digit_count  = 0 # mean relative to total number of responses
        flow.udps.dst2src_mean_dns_response_alpha_count  = 0 
        flow.udps.dst2src_mean_dns_response_hypens_count = 0 
        flow.udps.dst2src_mean_dns_response_dots_count   = 0 
        flow.udps.dst2src_mean_dns_response_ip_count     = 0 
        flow.udps.dst2src_mean_dns_response_ttls_s             = None
        flow.udps.dst2src_median_dns_response_ttls_s           = None
        flow.udps.dst2src_stdev_dns_response_ttls_s            = None
        flow.udps.dst2src_variance_dns_response_ttls_s         = None
        flow.udps.dst2src_coeff_of_var_dns_response_ttls_s     = None
        flow.udps.dst2src_skew_from_median_dns_response_ttls_s = None
        flow.udps.dst2src_min_dns_response_ttls_s              = None
        flow.udps.dst2src_max_dns_response_ttls_s              = None
                
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if flow.protocol == 17: # 17=UDP, 6=TCP
            ip_packet = IP(packet.ip_packet)
            if ip_packet[IP].version == 6:
                ip_packet = IPv6(packet.ip_packet)
            if DNS in ip_packet:
                dns_packet = ip_packet[DNS]
                if packet.direction == 0: # src2dst
                    flow.udps.src2dst_dns_packets  += 1
                    flow.udps.src2dst_dns_qd_count += dns_packet.qdcount
                    flow.udps.src2dst_dns_an_count += dns_packet.ancount
                    flow.udps.src2dst_dns_ns_count += dns_packet.nscount
                    flow.udps.src2dst_dns_ar_count += dns_packet.arcount
                    if dns_packet.qr == 0: # DNS query
                        flow.udps.src2dst_dns_queries += 1
                    elif dns_packet.qr == 1: # DNS response
                        flow.udps.src2dst_dns_responses += 1
                        flow.udps.src2dst_dns_response_digit_count  = self._count_rrname_char_of(dns_packet, lambda c: c.isdigit()) # counting digits 
                        flow.udps.src2dst_dns_response_alpha_count  = self._count_rrname_char_of(dns_packet, lambda c: c.isalpha()) # counting letters
                        flow.udps.src2dst_dns_response_hypens_count = self._count_rrname_char_of(dns_packet, lambda c: c == '-') # counting hypens
                        flow.udps.src2dst_dns_response_dots_count   = self._count_rrname_char_of(dns_packet, lambda c: c == '.') # counting dots
                        flow.udps.src2dst_dns_response_ip_count     = self._count_number_of_ip(dns_packet)
                        flow.udps.src2dst_dns_response_ttls_s.extend(self._ttl_values(dns_packet))
                        
                else: # dst2src
                    flow.udps.dst2src_dns_packets  += 1
                    flow.udps.dst2src_dns_qd_count += dns_packet.qdcount
                    flow.udps.dst2src_dns_an_count += dns_packet.ancount
                    flow.udps.dst2src_dns_ns_count += dns_packet.nscount
                    flow.udps.dst2src_dns_ar_count += dns_packet.arcount
                    if dns_packet.qr == 0: # DNS query
                        flow.udps.dst2src_dns_queries += 1
                    elif dns_packet.qr == 1: # DNS response
                        flow.udps.dst2src_dns_responses += 1
                        flow.udps.dst2src_dns_response_digit_count  = self._count_rrname_char_of(dns_packet, lambda c: c.isdigit()) # counting digits 
                        flow.udps.dst2src_dns_response_alpha_count  = self._count_rrname_char_of(dns_packet, lambda c: c.isalpha()) # counting letters
                        flow.udps.dst2src_dns_response_hypens_count = self._count_rrname_char_of(dns_packet, lambda c: c == '-') # counting hypens
                        flow.udps.dst2src_dns_response_dots_count   = self._count_rrname_char_of(dns_packet, lambda c: c == '.') # counting dots
                        flow.udps.dst2src_dns_response_ip_count     = self._count_number_of_ip(dns_packet)
                        flow.udps.dst2src_dns_response_ttls_s.extend(self._ttl_values(dns_packet))
                        
                # compute/update bidirectional by merging both directions.
                flow.udps.bidirectional_dns_packets   = flow.udps.src2dst_dns_packets   + flow.udps.dst2src_dns_packets
                flow.udps.bidirectional_dns_queries   = flow.udps.src2dst_dns_queries   + flow.udps.dst2src_dns_queries
                flow.udps.bidirectional_dns_responses = flow.udps.src2dst_dns_responses + flow.udps.dst2src_dns_responses 
                flow.udps.bidirectional_dns_qd_count  = flow.udps.src2dst_dns_qd_count  + flow.udps.dst2src_dns_qd_count
                flow.udps.bidirectional_dns_an_count  = flow.udps.src2dst_dns_an_count  + flow.udps.dst2src_dns_an_count
                flow.udps.bidirectional_dns_ns_count  = flow.udps.src2dst_dns_ns_count  + flow.udps.dst2src_dns_ns_count
                flow.udps.bidirectional_dns_ar_count  = flow.udps.src2dst_dns_ar_count  + flow.udps.dst2src_dns_ar_count
                flow.udps.bidirectional_dns_response_digit_count  = flow.udps.src2dst_dns_response_digit_count  + flow.udps.dst2src_dns_response_digit_count
                flow.udps.bidirectional_dns_response_alpha_count  = flow.udps.src2dst_dns_response_alpha_count  + flow.udps.dst2src_dns_response_alpha_count
                flow.udps.bidirectional_dns_response_hypens_count = flow.udps.src2dst_dns_response_hypens_count + flow.udps.dst2src_dns_response_hypens_count
                flow.udps.bidirectional_dns_response_dots_count   = flow.udps.src2dst_dns_response_dots_count   + flow.udps.dst2src_dns_response_dots_count
                flow.udps.bidirectional_dns_response_ip_count     = flow.udps.src2dst_dns_response_ip_count     + flow.udps.dst2src_dns_response_ip_count 
                flow.udps.bidirectional_dns_response_ttls_s       = flow.udps.src2dst_dns_response_ttls_s       + flow.udps.dst2src_dns_response_ttls_s
    
    def on_expire(self, flow):
        # TTLS
        # src -> dst
        if len(flow.udps.src2dst_dns_response_ttls_s) > 0:
            src2dst_stats = IterableStats(flow.udps.src2dst_dns_response_ttls_s)
            flow.udps.src2dst_mean_dns_response_ttls_s         = src2dst_stats.average()
            flow.udps.src2dst_median_dns_response_ttls_s       = src2dst_stats.median()
            flow.udps.src2dst_stdev_dns_response_ttls_s        = src2dst_stats.std_deviation()
            flow.udps.src2dst_variance_dns_response_ttls_s     = src2dst_stats.variance()
            flow.udps.src2dst_coeff_of_var_dns_response_ttls_s = src2dst_stats.coeff_of_variation()
            flow.udps.src2dst_skew_from_median_dns_response_ttls_s = src2dst_stats.skew_from_median()
            flow.udps.src2dst_min_dns_response_ttls_s          = src2dst_stats.min()
            flow.udps.src2dst_max_dns_response_ttls_s          = src2dst_stats.max()
        else: 
            flow.udps.src2dst_dns_response_ttls_s = None
        # dst -> src
        if len(flow.udps.dst2src_dns_response_ttls_s) > 0:
            dst2src_stats = IterableStats(flow.udps.dst2src_dns_response_ttls_s)
            flow.udps.dst2src_mean_dns_response_ttls_s             = dst2src_stats.average()
            flow.udps.dst2src_median_dns_response_ttls_s           = dst2src_stats.median()
            flow.udps.dst2src_stdev_dns_response_ttls_s            = dst2src_stats.std_deviation()
            flow.udps.dst2src_variance_dns_response_ttls_s         = dst2src_stats.variance()
            flow.udps.dst2src_coeff_of_var_dns_response_ttls_s     = dst2src_stats.coeff_of_variation()
            flow.udps.dst2src_skew_from_median_dns_response_ttls_s = dst2src_stats.skew_from_median()
            flow.udps.dst2src_min_dns_response_ttls_s              = dst2src_stats.min()
            flow.udps.dst2src_max_dns_response_ttls_s              = dst2src_stats.max()
        else: 
            flow.udps.dst2src_dns_response_ttls_s = None
        # bidirectional
        if len(flow.udps.bidirectional_dns_response_ttls_s) > 0:
            bi_stats = IterableStats(flow.udps.bidirectional_dns_response_ttls_s)
            flow.udps.bidirectional_mean_dns_response_ttls_s             = bi_stats.average()
            flow.udps.bidirectional_median_dns_response_ttls_s           = bi_stats.median()
            flow.udps.bidirectional_stdev_dns_response_ttls_s            = bi_stats.std_deviation()
            flow.udps.bidirectional_variance_dns_response_ttls_s         = bi_stats.variance()
            flow.udps.bidirectional_coeff_of_var_dns_response_ttls_s     = bi_stats.coeff_of_variation()
            flow.udps.bidirectional_skew_from_median_dns_response_ttls_s = bi_stats.skew_from_median()
            flow.udps.bidirectional_min_dns_response_ttls_s              = bi_stats.min()
            flow.udps.bidirectional_max_dns_response_ttls_s              = bi_stats.max()
        else: 
            flow.udps.bidirectional_dns_response_ttls_s = None
            
        # Char/Digits/Dots/Hypens
        # src -> dst
        if flow.udps.src2dst_dns_responses > 0:
            flow.udps.src2dst_mean_dns_response_digit_count  = flow.udps.src2dst_dns_response_digit_count  / flow.udps.src2dst_dns_responses
            flow.udps.src2dst_mean_dns_response_alpha_count  = flow.udps.src2dst_dns_response_alpha_count  / flow.udps.src2dst_dns_responses
            flow.udps.src2dst_mean_dns_response_dots_count   = flow.udps.src2dst_dns_response_dots_count   / flow.udps.src2dst_dns_responses
            flow.udps.src2dst_mean_dns_response_ip_count     = flow.udps.src2dst_dns_response_ip_count     / flow.udps.src2dst_dns_responses
            flow.udps.src2dst_mean_dns_response_hypens_count = flow.udps.src2dst_dns_response_hypens_count / flow.udps.src2dst_dns_responses
        # dst -> src
        if flow.udps.dst2src_dns_responses > 0:
            flow.udps.dst2src_mean_dns_response_digit_count  = flow.udps.dst2src_dns_response_digit_count  / flow.udps.dst2src_dns_responses
            flow.udps.dst2src_mean_dns_response_alpha_count  = flow.udps.dst2src_dns_response_alpha_count  / flow.udps.dst2src_dns_responses
            flow.udps.dst2src_mean_dns_response_dots_count   = flow.udps.dst2src_dns_response_dots_count   / flow.udps.dst2src_dns_responses
            flow.udps.dst2src_mean_dns_response_ip_count     = flow.udps.dst2src_dns_response_ip_count     / flow.udps.dst2src_dns_responses
            flow.udps.dst2src_mean_dns_response_hypens_count = flow.udps.dst2src_dns_response_hypens_count / flow.udps.dst2src_dns_responses
        # bidir
        if flow.udps.bidirectional_dns_responses > 0:
            flow.udps.bidirectional_mean_dns_response_digit_count  = flow.udps.bidirectional_dns_response_digit_count  / flow.udps.bidirectional_dns_responses
            flow.udps.bidirectional_mean_dns_response_alpha_count  = flow.udps.bidirectional_dns_response_alpha_count  / flow.udps.bidirectional_dns_responses
            flow.udps.bidirectional_mean_dns_response_dots_count   = flow.udps.bidirectional_dns_response_dots_count   / flow.udps.bidirectional_dns_responses
            flow.udps.bidirectional_mean_dns_response_ip_count     = flow.udps.bidirectional_dns_response_ip_count     / flow.udps.bidirectional_dns_responses
            flow.udps.bidirectional_mean_dns_response_hypens_count = flow.udps.bidirectional_dns_response_hypens_count / flow.udps.bidirectional_dns_responses
    
    def _count_rrname_char_of(self, dns_res_packet, count_func):
        resource_record_names = []
        for i in range(dns_res_packet.ancount):
            dns_ans_record = dns_res_packet.an[i]
            resource_record_names += str(dns_ans_record.rrname)[2:-2]
        total_digit_count  = sum(count_func(c) for c in resource_record_names)
        
        return total_digit_count
    
    def _count_number_of_ip(self, dns_res_packet):
        # Note: not tested
        '''
            Records: 
                A=1 (0x01)
                AAAA=28 (0x1c)
        '''
        ip_count = 0
        for i in range(dns_res_packet.ancount):
            dns_ans_record = dns_res_packet.an[i]
            if dns_ans_record.type in [1, 28]:
                ip_count += 1
                
        return ip_count
        
    def _ttl_values(self, dns_res_packet):
        ttls = []
        for i in range(dns_res_packet.ancount):
            dns_ans_record = dns_res_packet.an[i]
            ttls.append(dns_ans_record.ttl)
                
        return ttls
    
            
            
            




from stats.stats import IterableStats
from nfstream import NFPlugin


class PacketRelativeTime(NFPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''
        # bidirectional
        flow.udps.bidirectional_packet_relative_times = list()
        flow.udps.bidirectional_min_packet_relative_times              = 0
        flow.udps.bidirectional_max_packet_relative_times              = 0
        flow.udps.bidirectional_mean_packet_relative_times             = 0
        flow.udps.bidirectional_stddev_packet_relative_times           = None
        flow.udps.bidirectional_variance_packet_relative_times         = None
        flow.udps.bidirectional_coeff_of_var_packet_relative_times     = None
        flow.udps.bidirectional_skew_from_median_packet_relative_times = None
        # src -> dst
        flow.udps.src2dst_packet_relative_times = list()
        flow.udps.src2dst_min_packet_relative_times              = 0
        flow.udps.src2dst_max_packet_relative_times              = 0
        flow.udps.src2dst_mean_packet_relative_times             = 0
        flow.udps.src2dst_stddev_packet_relative_times           = None
        flow.udps.src2dst_variance_packet_relative_times         = None
        flow.udps.src2dst_coeff_of_var_packet_relative_times     = None
        flow.udps.src2dst_skew_from_median_packet_relative_times = None
        # dst -> src
        flow.udps.dst2src_packet_relative_times = list()
        flow.udps.dst2src_min_packet_relative_times              = 0
        flow.udps.dst2src_max_packet_relative_times              = 0
        flow.udps.dst2src_mean_packet_relative_times             = 0
        flow.udps.dst2src_stddev_packet_relative_times           = None
        flow.udps.dst2src_variance_packet_relative_times         = None
        flow.udps.dst2src_coeff_of_var_packet_relative_times     = None
        flow.udps.dst2src_skew_from_median_packet_relative_times = None

        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if packet.direction == 0: # src -> dst
            flow.udps.src2dst_packet_relative_times.append(packet.time
                                    - flow.bidirectional_first_seen_ms)
        elif packet.direction == 1:
            flow.udps.dst2src_packet_relative_times.append(packet.time
                                    - flow.bidirectional_first_seen_ms)
            
        flow.udps.bidirectional_packet_relative_times.append(packet.time
                                    - flow.bidirectional_first_seen_ms)

        
    def on_expire(self, flow):
        # bidirectional
        stats = IterableStats(flow.udps.bidirectional_packet_relative_times)
        flow.udps.bidirectional_min_packet_relative_times              = stats.min()
        flow.udps.bidirectional_max_packet_relative_times              = stats.max()
        flow.udps.bidirectional_mean_packet_relative_times             = stats.average()
        flow.udps.bidirectional_stddev_packet_relative_times           = stats.std_deviation()
        flow.udps.bidirectional_variance_packet_relative_times         = stats.variance()
        flow.udps.bidirectional_coeff_of_var_packet_relative_times     = stats.coeff_of_variation()
        flow.udps.bidirectional_skew_from_median_packet_relative_times = stats.skew_from_median()
        # src -> dst
        stats = IterableStats(flow.udps.src2dst_packet_relative_times)
        flow.udps.src2dst_min_packet_relative_times              = stats.min()
        flow.udps.src2dst_max_packet_relative_times              = stats.max()
        flow.udps.src2dst_mean_packet_relative_times             = stats.average()
        flow.udps.src2dst_stddev_packet_relative_times           = stats.std_deviation()
        flow.udps.src2dst_variance_packet_relative_times         = stats.variance()
        flow.udps.src2dst_coeff_of_var_packet_relative_times     = stats.coeff_of_variation()
        flow.udps.src2dst_skew_from_median_packet_relative_times = stats.skew_from_median()
        # dst -> src
        stats = IterableStats(flow.udps.dst2src_packet_relative_times)
        flow.udps.dst2src_min_packet_relative_times              = stats.min()
        flow.udps.dst2src_max_packet_relative_times              = stats.max()
        flow.udps.dst2src_mean_packet_relative_times             = stats.average()
        flow.udps.dst2src_stddev_packet_relative_times           = stats.std_deviation()
        flow.udps.dst2src_variance_packet_relative_times         = stats.variance()
        flow.udps.dst2src_coeff_of_var_packet_relative_times     = stats.coeff_of_variation()
        flow.udps.dst2src_skew_from_median_packet_relative_times = stats.skew_from_median()



import numpy as np
import math 


class Stats: # INTERFACE
    def average(self):
        pass

    def median(self):
        pass

    def mode(self):
        pass 

    def variance(self):
        pass
    
    def std_deviation(self):
        pass

    def coeff_of_variation(self):
        pass

    def skew_from_median(self):
        pass

    def skew_from_mode(self):
        pass 
    
    def min(self):
        pass
    
    def max(self):
        pass


class StatsCollection(Stats): 
    def __init__(self, stats) -> None:
        self.inner_collection = stats

    def average(self):
        avgs = np.empty(len(self.inner_collection))
        for i, stats in enumerate(self.inner_collection):
            avgs[i] = stats.average()

        return avgs

    def median(self):
        pass

    def mode(self):
        pass 

    def variance(self):
        pass
    
    def std_deviation(self):
        pass

    def coeff_of_variation(self):
        pass

    def skew_from_median(self):
        pass

    def skew_from_mode(self):
        pass 
    
    def min(self) -> float:
        return np.min(self.values)
    
    def max(self) -> float:
        return np.max(self.values)

class WeightedIterableStats(Stats):
    '''
        Works only on iterables such as lists.
        Assumes that:
        1. the values and weights has matching indices,
        2. that the values are ordered/sorted from low to high.
        3. the weights are integers. (frequencies) for median calc
                                                '''
    def __init__(self, values, weights) -> None:
        self.values = values
        self.weights = weights

    def average(self) -> float:
        if np.sum(self.weights) == 0:
            return 0
        return np.average(self.values, weights=self.weights)
    
    def median(self) -> float:
        total_weight = np.sum(self.weights)
        half_total_weight = total_weight/2
        is_weight_even = (total_weight % 2) == 0
        accumulated_weight = 0
        for i, weight in enumerate(self.weights):
            accumulated_weight += weight
            if is_weight_even:
                if accumulated_weight == half_total_weight-0.5:
                    return (self.values[i-1] + self.values[i])/2
                elif accumulated_weight == half_total_weight+0.5:
                    return self.values[i]
            if accumulated_weight >= half_total_weight:
                return self.values[i]
            
        raise Exception('Logical error.')

    def mode(self) -> float:
        raise Exception('Not implemented') 

    def variance(self) -> float:
        if np.sum(self.weights) == 0:
            return 0
        return np.average((self.values-self.average())**2, weights=self.weights)
    
    def std_deviation(self) -> float:
        return math.sqrt(self.variance())

    def coeff_of_variation(self) -> float:
        if self.average() != 0:
            return self.std_deviation() / self.average()
        else:
            return 0

    def skew_from_median(self) -> float:
        #  Skew = 3 * (Mean – Median) / Standard Deviation
        if self.std_deviation() != 0:
            return 3 * (self.average() - self.median()) / self.std_deviation()
        else:
            return 0

    def skew_from_mode(self) -> float:
        #  Skew =  (Mean – Mode) / Standard Deviation
        if self.std_deviation() != 0:
            return (self.average() - self.mode()) / self.std_deviation()
        else:
            return 0
    
    def min(self):
        if np.sum(self.weights) == 0:
            return 0
        else:    
            return self.values[np.argmin(self.weights)]
    
    def max(self):
        if np.sum(self.weights) == 0:
            return 0
        else:
            return self.values[np.argmax(self.weights)]



class IterableStats(Stats):
    '''
        Works only on iterables such as lists.
                                                '''
    def __init__(self, values) -> None:
        self.values = values

    def average(self) -> float:
        if len(self.values) != 0:
            return np.average(self.values)
        else:
            return None
    
    def median(self) -> float:
        if len(self.values) != 0:
            return np.median(self.values)
        else:
            return None

    def mode(self) -> float:
        raise Exception('Not implemented') 

    def variance(self) -> float:
        if len(self.values) >= 2:
            return np.var(self.values)
        else:
            return None
    
    def std_deviation(self) -> float:
        if len(self.values) >= 2:
            return np.std(self.values)
        else:
            return None
        

    def coeff_of_variation(self) -> float:
        avg = self.average()
        std = self.std_deviation()
        if avg not in {0, None} and std is not None:
            return std / avg
        else:
            return None

    def skew_from_median(self) -> float:
        #  Skew = 3 * (Mean – Median) / Standard Deviation
        std = self.std_deviation()
        if std not in {0, None}:
            return 3 * (self.average() - self.median()) / std
        else:
            return None

    def skew_from_mode(self) -> float:
        #  Skew =  (Mean – Mode) / Standard Deviation
        if self.std_deviation() not in {0, None}:
            return (self.average() - self.mode()) / self.std_deviation()
        else:
            return None
    
    def min(self):
        if len(self.values) != 0:
            return np.min(self.values)
        else:
            return None
    
    def max(self):
        if len(self.values) != 0:
            return np.max(self.values)
        else:
            return None

       


from nfstream import NFPlugin
import numpy as np 


class MostFreqPayloadLenRatio(NFPlugin):
    '''
        The ratio of number of packets with most freq payload len for direction X to the the total number of packets
        in direction X,
        for each direction x in {src2dst, dst2src}.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''

        flow.udps.src2dst_most_freq_payload_ratio = 0
        flow.udps.src2dst_most_freq_payload_len   = None
        flow.udps.src2dst_payload_freq = dict()
        flow.udps.dst2src_most_freq_payload_ratio = 0
        flow.udps.dst2src_most_freq_payload_len   = None
        flow.udps.dst2src_payload_freq = dict()

        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if packet.direction == 0: # src2dst
            if packet.payload_size not in flow.udps.src2dst_payload_freq:
                flow.udps.src2dst_payload_freq[packet.payload_size] = 0
            flow.udps.src2dst_payload_freq[packet.payload_size] +=1
        else:
            if packet.payload_size not in flow.udps.dst2src_payload_freq:
                flow.udps.dst2src_payload_freq[packet.payload_size] = 0
            flow.udps.dst2src_payload_freq[packet.payload_size] +=1

    def on_expire(self, flow):
        if flow.src2dst_packets != 0:
            freq_dict = flow.udps.src2dst_payload_freq
            most_freq_payload_freq = max(freq_dict.values())
            flow.udps.src2dst_most_freq_payload_len   = list(freq_dict.keys())[np.argmax(list(freq_dict.values()))]
            flow.udps.src2dst_most_freq_payload_ratio = most_freq_payload_freq / flow.src2dst_packets
        if flow.dst2src_packets != 0:
            freq_dict = flow.udps.dst2src_payload_freq
            most_freq_payload_freq = max(freq_dict.values())
            flow.udps.dst2src_most_freq_payload_len   = list(freq_dict.keys())[np.argmax(list(freq_dict.values()))]
            flow.udps.dst2src_most_freq_payload_ratio = most_freq_payload_freq / flow.dst2src_packets



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # or any {'0', '1', '2'}
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.WARN)  # or any {DEBUG, INFO, WARN, ERROR, FATAL}
from tensorflow.keras.layers import *
from tensorflow.keras.models import *




class CustomDistiller(Model):
    def __init__(self, modalities=[], adapter_size=128, n_classes=[]) -> None:
        super(CustomDistiller, self).__init__()
        self.n_classes = n_classes
        self.modalities = modalities
        self.adapter_size = adapter_size
        self._validate()
        
        shared_representation = stack(
            [
                Concatenate()(
                    wrap_adapter_multi(modalities, adapter_size) 
                )
            ]
            + get_sr_layers(adapter_size)
        )
        
        outputs = []
        for n_class in n_classes:
            outputs.append(stack([shared_representation] + get_ts_layers(classes_count=n_class, adapter_size=adapter_size)))
        self.model = Model(
            name='Distiller',
            inputs=[modal.input for modal in modalities],
            outputs= outputs
        )


    def call(self, inputs, training=None):
        # See: https://www.tensorflow.org/guide/keras/custom_layers_and_models#the_model_class
        return self.model(inputs, training)


    def _validate(self):
        # modalities and n_classess array must not be empty .
        assert len(self.modalities) != 0
        assert len(self.n_classes) != 0
        
        # n_classes values must represent the number of classes in each task.
        # and therefore must be positive integers.
        for nclass in self.n_classes:
            assert nclass >= 1
        
        # adapter_size must be positive.
        assert self.adapter_size >= 1


    def get_model_for_pretraining(self, model):
        model_w_adapter = wrap_adapter(model, self.adapter_size)
        outputs = []
        for n_class in self.n_classes:
            outputs.append(stack([model_w_adapter, Dense(n_class, activation='softmax')]))
        return Model(
            name='pretraining_model',
            inputs=model.input,
            outputs=outputs
        )


    def freeze_for_finetuning(self):
        for modal in self.modalities:
            for layer in modal.layers:
                layer.trainable = False

  
    def unfreeze_for_finetuning(self):
        for modal in self.modalities:
            for layer in modal.layers:
                layer.trainable = True
        
        
    def fit(self, x, y, **kwargs):
        loss_fn = tf.keras.losses.CategoricalCrossentropy(from_logits=False)
        for features, modal in zip(x, self.modalities):
            print('##################### {} ##########################'.format(modal.name.upper()))
            pretraining_model = self.get_model_for_pretraining(modal)
            pretraining_model.compile(
                optimizer=tf.keras.optimizers.Adam(learning_rate=0.002),
                loss=loss_fn,
                metrics=[['accuracy']] * len(self.n_classes)
            ) 
            
            pretraining_model.fit(features, y, **kwargs)

        # FINE-TUNE
        print('##################### FINE-TUNING ##########################')
        self.freeze_for_finetuning()
        
        self.model.compile(
            optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
            loss=loss_fn,
            metrics=['accuracy']
        )

        self.model.fit(x,y, **kwargs)

        self.unfreeze_for_finetuning()



###################
# Model utilities #
###################


def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack


def get_adapter_layers(adapter_size):
    return [
        Dropout(0.2),
        Dense(adapter_size),
        ReLU()
    ]


def wrap_adapter(model, adapter_size):
    return stack([model.output, *get_adapter_layers(adapter_size)])


def wrap_adapter_multi(models, adapter_size):
    return [wrap_adapter(model, adapter_size) for model in models]


def get_sr_layers(adapter_size):
    # SR = Shared Representation
    return [
        Dropout(0.2),
        Dense(adapter_size),
        ReLU(),
        Dropout(0.2),
    ]


def get_ts_layers(classes_count, adapter_size):
    # TS = Task Specific
    return [
        Dense(adapter_size),
        ReLU(),
        Dropout(0.2),
        Dense(classes_count),
        Softmax()
    ]


from third_party.tshark.dataframe_utils import add_packet_clump_num_column, add_packet_directions_column, convert_to_correct_dtypes, drop_lengthless_tls_records, get_tls_clump_stats, get_tls_record_stats, groupby_biflows, merge_df_by_biflows, read_tshark_csv_output
from third_party.tshark.tls_stats import TLSRecordClumpStats, TLSRecordStats
import json
from third_party.tshark.tshark import run_tshark_tls
from datetime import datetime
import os

'''
Meant to work under linux system.
Tested with WSL ubuntu 18.04.
TShark (Wireshark 3.4.1+)
'''

def extract_tls_features_and_merge(df_nfstream, pcap_filepath, config_filepath,):
    config = _read_config_file(config_filepath)
    tls_features_per_session = _extract_tls_features(pcap_filepath, config['tshark_location'])
    print('Read and merge dataframe of TShark with NFStream', datetime.now().strftime("%H:%M:%S"))
    if tls_features_per_session is None:
        col_names = TLSRecordStats.available_feature_names() + TLSRecordClumpStats.available_feature_names()
        df = df_nfstream.copy()
        df[col_names] = None
    else:
        df = merge_df_by_biflows(tls_features_per_session, df_nfstream)
    print('Exiting 3rd-party: TShark for TLS...', datetime.now().strftime("%H:%M:%S"))
    return df


def _extract_tls_features(pcap_filepath, tshark_location):
    print('Running 3rd-party: Tshark for TLS...', datetime.now().strftime("%H:%M:%S"))
    pcap_file = pcap_filepath
    tshark_output_filepath = 'tshark_outfile.csv'
    run_tshark_tls(pcap_file, tshark_output_filepath, tshark_location)

    print('Reading Tshark CSV output as Pandas Dataframe', datetime.now().strftime("%H:%M:%S"))
    df = read_tshark_csv_output(tshark_output_filepath)
    df = drop_lengthless_tls_records(df)
    if df.empty:
        tls_features_per_session = None
    else:   
        df = convert_to_correct_dtypes(df)
        df = add_packet_directions_column(df)
        df = add_packet_clump_num_column(df)
        record_tls_per_session = get_tls_record_stats(df)
        clump_stats_per_session = get_tls_clump_stats(df)
        tls_features_per_session = record_tls_per_session.join(clump_stats_per_session)

    os.remove(tshark_output_filepath)

    return tls_features_per_session


def _read_config_file(config_filepath):
    print('Reading Configuration...', datetime.now().strftime("%H:%M:%S"))
    print('Configuration at', config_filepath)
    with open(config_filepath) as json_config:
        config = json.load(json_config)
        return config




from scapy.all import load_layer
load_layer("tls")

class TLSReassemble:
    ''' 
        Assumes that packets are received in order.
        For out of order packets, need to reorder TCP stream by
        seq/ack.

    '''
    def __init__(self) -> None:
        self.accumulated = 0
        self.last_record_size = 0
        self.last_record_received = None
    
    def process(self, tcp_payload):
        records = list()
        if self.accumulated == 0:
            tls = TLS(tcp_payload)
            payload_size = len(tcp_payload)
            accumulated_payload_size = 0
            i = 0
            while accumulated_payload_size < payload_size:
                record = tls[i]
                record_size = record.len
                received_data_size = record.deciphered_len
                accumulated_payload_size += received_data_size + 5 # 5 is TLS record header len 

                if record_size > received_data_size:
                    self.accumulated += received_data_size
                    self.last_record_size = record_size
                    self.last_record_received = record
                else:
                    records.append(tls[i])
                
                i += 2
        else:
            completing_size = self.last_record_size - self.accumulated
            completing_payload = tcp_payload[:completing_size]
            self.last_record_received.msg[0].data += completing_payload
            records.append(self.last_record_received)
            self.accumulated = 0
            self.last_record_size = 0
            self.last_record_received_data = None

            return records + self.process(tcp_payload[completing_size:])

        return records

            
            

            

            


from nfstream import NFPlugin

class RecvSentPacketRatio(NFPlugin):
    '''
        The ratio of the number of received packets to the number of sent packets.
        = recv/sent.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def on_expire(self, flow):
        flow.udps.sent_recv_packet_ratio =  flow.dst2src_packets / flow.src2dst_packets




from nfstream import NFPlugin
import pyasn
import functools
import pandas as pd
from os import path

class ASNInfo(NFPlugin):
    '''
    Description:
        Extracts ASN related info from source and destination ip addresses.
        By using databases along with 'pyasn' Python module.
        DB required: 
            1. pyasn.db or equavalent, please see 'IPASN Data Files' section in: 
            https://github.com/hadiasghari/pyasn to aquire it.
            2. TSV (tab separated values) DB file of contextual info such as from https://iptoasn.com
            
    Features:
        asn_number
        asn_country_code
        asn_description
        
    Feature prefixes:
        src_
        dst_
    '''
    def __init__(self, pyasn_context_file=path.join(path.dirname(__file__), '../tools/pyasn.db'), 
                as_contextual_file=path.join(path.dirname(__file__),'../tools/ip2asn-v4.tsv'), **kwargs):
        super().__init__(**kwargs)
        self.pyasn_contextual_data = pyasn.pyasn(pyasn_context_file)
        self.as_contextual_data = pd.read_csv(as_contextual_file, 
                                            sep='\t', 
                                            names=[
                                                'range_start',
                                                'range_end', 
                                                'AS_number', 
                                                'country_code', 
                                                'AS_description'
                                            ])

    def on_init(self, packet, flow):
        flow.udps.src_asn_number       = None
        flow.udps.src_asn_country_code = None
        flow.udps.src_asn_description  = None
        flow.udps.dst_asn_number       = None
        flow.udps.dst_asn_country_code = None
        flow.udps.dst_asn_description  = None
        # src asn info
        src_asn_info = self.get_asn_info(flow.src_ip)
        if src_asn_info is not None:
            flow.udps.src_asn_number       = src_asn_info['AS_number']
            flow.udps.src_asn_country_code = src_asn_info['country_code']
            flow.udps.src_asn_description  = src_asn_info['AS_description']
        # dst asn info
        dst_asn_info = self.get_asn_info(flow.dst_ip)
        if dst_asn_info is not None:
            flow.udps.dst_asn_number       = dst_asn_info['AS_number']
            flow.udps.dst_asn_country_code = dst_asn_info['country_code']
            flow.udps.dst_asn_description  = dst_asn_info['AS_description']

    # extracted with modification from https://github.com/cisco/mercury/blob/c8ad79d56959ba1092e686b23c1b6f60961600c7/src/python-inference/tls_fingerprint_min.pyx
    MAX_CACHED_RESULTS = 2**16
    @functools.lru_cache(maxsize=MAX_CACHED_RESULTS)
    def get_asn_info(self, ip_addr: str):
        try:
            asn_n,_ = self.pyasn_contextual_data.lookup(ip_addr)
            if asn_n != None:
                return self.as_contextual_data[self.as_contextual_data['AS_number'] == asn_n].iloc[0]
        except:
            pass

        return None


from nfstream import NFPlugin
import numpy as np
from PIL import Image
import os


class GrayPic1(NFPlugin):
    '''
    Description:
        Produces two gray-scaled image/picture, first of size 16x40 (IAT to packet-sizes),
        and the second of size 30x40 (burst/clump time to burst/clump size in bytes).
        data is extracted from the 1024-packet windows of bi-flow (default), this number can be configured.
    
    Paper:
        "Fingerprinting encrypted network traffic types using machine learning"
    By:
        Sam Leroux, 
        Steven Bohez, 
        Pieter-Jan Maenhaut, 
        Nathan Meheus, 
        Pieter Simoens, 
        Bart Dhoedt.
        - Department of Information Technology, IDLab, Ghent University - imec
    Source:
        https://biblio.ugent.be/publication/8559975/file/8559977.pdf
    '''

    def __init__(self, save_path, n_first_packets=1024, **kwargs):
        super().__init__(**kwargs)
        self.save_path = save_path
        self.n_first_packets = n_first_packets

    def on_init(self, packet, flow):
        flow.udps.graypic1_iat2packetsizes     = np.zeros((40, 16), dtype=np.uint32)
        flow.udps.graypic1_bursttime2burstsize = np.zeros((40, 30), dtype=np.uint32)
        flow.udps.graypic1_previous_packet_time = packet.time
        flow.udps.graypic1_packets = 0
        flow.udps.graypic1_pic_count = 0
        flow.udps.graypic1_burst_size = 0 # in bytes
        flow.udps.graypic1_burst_direction = packet.direction 
        flow.udps.graypic1_burst_start_time = packet.time
        flow.udps.graypic1_burst_end_time = packet.time
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        flow.udps.graypic1_packets += 1
        self._handle_iat2packetsizes(packet, flow)
        self._handle_bursttime2burstsize(packet, flow)
        flow.udps.graypic1_previous_packet_time = packet.time
        
        if flow.udps.graypic1_packets == self.n_first_packets:
            self._handle_iat2packetsizes_finalize(packet, flow)
            self._handle_bursttime2burstsize_finalize(packet, flow)
            flow.udps.graypic1_packets = 0
            flow.udps.graypic1_pic_count += 1
        
    def on_expire(self, flow):
        # cleanup
        del flow.udps.graypic1_iat2packetsizes     
        del flow.udps.graypic1_bursttime2burstsize 
        del flow.udps.graypic1_previous_packet_time
        del flow.udps.graypic1_packets
        del flow.udps.graypic1_pic_count
        del flow.udps.graypic1_burst_size
        del flow.udps.graypic1_burst_direction 
        del flow.udps.graypic1_burst_start_time
        del flow.udps.graypic1_burst_end_time

    def _handle_iat2packetsizes(self, packet, flow):
        scaled_size = int(np.log(packet.ip_size)) * (1 if packet.direction==0 else -1)
        if packet.time - flow.udps.graypic1_previous_packet_time != 0:
            scaled_interarrival_time = int(np.log(packet.time - flow.udps.graypic1_previous_packet_time)) * (1 if packet.direction==0 else -1)
        else:
            scaled_interarrival_time = 0
        flow.udps.graypic1_iat2packetsizes[20+scaled_interarrival_time, 8+scaled_size] += 1
        
       
    def _handle_iat2packetsizes_finalize(self, packet, flow):
        # save and stuff
        flow.udps.graypic1_iat2packetsizes = ((flow.udps.graypic1_iat2packetsizes/np.sum(flow.udps.graypic1_iat2packetsizes))*255).astype(np.uint8)
        flow.udps.graypic1_iat2packetsizes = np.flip(flow.udps.graypic1_iat2packetsizes, axis=0)
        flow.udps.graypic1_iat2packetsizes = 255-flow.udps.graypic1_iat2packetsizes # for clarity/debug
        img = Image.fromarray(flow.udps.graypic1_iat2packetsizes, 'L') # L = 8bit black and white (gray scale)
        img.save(os.path.join(self.save_path,
                        '-'.join(['graypic1',str(flow.udps.graypic1_pic_count), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol)])
                        + '.png'))
        print('saved:', os.path.join(self.save_path,
                        '-'.join(['graypic1',str(flow.udps.graypic1_pic_count), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol)])
                        + '.png'))
        
        flow.udps.graypic1_iat2packetsizes = np.zeros((40, 16), dtype=np.uint32)
            
    def _handle_bursttime2burstsize(self, packet, flow):
        if packet.payload_size == 0: # skip ACK packets and empty payload packets
            return
        
        if flow.udps.graypic1_burst_direction == -1:
            flow.udps.graypic1_burst_direction = packet.direction 
            flow.udps.graypic1_burst_start_time = packet.time
            
        if packet.direction != flow.udps.graypic1_burst_direction:
            if flow.udps.graypic1_burst_end_time - flow.udps.graypic1_burst_start_time != 0:
                scaled_burst_duration = int(np.log(flow.udps.graypic1_burst_end_time - flow.udps.graypic1_burst_start_time)) * (1 if packet.direction==0 else -1) # ms 
            else:
                scaled_burst_duration = 0
            if flow.udps.graypic1_burst_size != 0:
                scaled_burst_size = int(np.log(flow.udps.graypic1_burst_size)) * (1 if packet.direction==0 else -1) # in bytes
            else:
                scaled_burst_size = 0
            flow.udps.graypic1_bursttime2burstsize[20+scaled_burst_duration, 15+scaled_burst_size] += 1
            
            flow.udps.graypic1_burst_start_time = packet.time
            flow.udps.graypic1_burst_end_time   = packet.time
            flow.udps.graypic1_burst_size       = packet.ip_size
            flow.udps.graypic1_burst_direction  = packet.direction
        else:
            flow.udps.graypic1_burst_end_time = packet.time
            flow.udps.graypic1_burst_size    += packet.ip_size
        
              
    def _handle_bursttime2burstsize_finalize(self, packet, flow):
        # save and stuff
        flow.udps.graypic1_bursttime2burstsize = ((flow.udps.graypic1_bursttime2burstsize/np.sum(flow.udps.graypic1_bursttime2burstsize))*255).astype(np.uint8)
        flow.udps.graypic1_bursttime2burstsize = np.flip(flow.udps.graypic1_bursttime2burstsize, axis=0)
        flow.udps.graypic1_bursttime2burstsize = 255-flow.udps.graypic1_bursttime2burstsize # for clarity/debug
        img = Image.fromarray(flow.udps.graypic1_bursttime2burstsize, 'L') # L = 8bit black and white (gray scale)
        img.save(os.path.join(self.save_path,
                        '-'.join(['graypic1', 'burst',str(flow.udps.graypic1_pic_count), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol)])
                        + '.png'))
        print('saved:', os.path.join(self.save_path,
                        '-'.join(['graypic1', 'burst',str(flow.udps.graypic1_pic_count), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol)])
                        + '.png'))
        
       
        flow.udps.graypic1_bursttime2burstsize = np.zeros((40, 30), dtype=np.uint32)
        flow.udps.graypic1_burst_size = 0 # in bytes
        flow.udps.graypic1_burst_direction = -1


from nfstream import NFPlugin


class SmallPacketPayloadRatio(NFPlugin):
    '''
        The ratio of number of small packets for direction X to the the total number of packets
        in direction X,
        for each direction x in {src2dst, dst2src}.
    '''
    def __init__(self, small_size = 32, **kwargs):
        '''
            small_size: the size (in bytes) of the payload that
            a packet is considered to be small (i.e, payload size < small_size).
        '''
        super().__init__(**kwargs)
        self.small_size = small_size

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''
        flow.udps.src2dst_small_packet_payload_packets = 0
        flow.udps.src2dst_small_packet_payload_ratio   = 0
        flow.udps.dst2src_small_packet_payload_packets = 0
        flow.udps.dst2src_small_packet_payload_ratio   = 0
        

        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if packet.payload_size < self.small_size:
            if packet.direction == 0: # src2dst
                flow.udps.src2dst_small_packet_payload_packets += 1
            else:
                flow.udps.dst2src_small_packet_payload_packets += 1

    def on_expire(self, flow):
        if flow.src2dst_packets != 0:
            flow.udps.src2dst_small_packet_payload_ratio = (flow.udps.src2dst_small_packet_payload_packets 
                                                            / flow.src2dst_packets)
        if flow.dst2src_packets != 0:
            flow.udps.dst2src_small_packet_payload_ratio = (flow.udps.dst2src_small_packet_payload_packets
                                                            / flow.dst2src_packets)



import glob
import os
from flowpic2 import FlowPic2
from nfstream import NFStreamer
from datetime import datetime


def get_last_dir_in_path(dir_path):
    '''
    General Purpose
    '''
    return os.path.basename(os.path.normpath(dir_path))

def current_time():
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    return f'[{current_time}]'

print(current_time(), 'Starting...')
for fn in glob.glob(r'flowpic_extracted_flows/*/*.pcap'):
    print(current_time(), 'input', fn)
    output_dirpath = os.path.join(os.path.dirname(fn), 'output')
    print(current_time(), 'output', output_dirpath)
    try:
        os.mkdir(output_dirpath)
    except:
        print(current_time(), 'Directory creation: FAILED!', output_dirpath)
    
    streamer = NFStreamer(source=fn,
                        decode_tunnels=True,
                        bpf_filter=None,
                        promiscuous_mode=True,
                        snapshot_length=1536,
                        idle_timeout=99999999,
                        active_timeout=99999999,
                        accounting_mode=3,
                        udps=[
                        #FlowPic2019('./temp', 0,  flow_active_time=60)
                        FlowPic2(output_dirpath, time_per_subflow=60)
                        #GrayPic1('./temp')
                        ],
                        n_dissections=20,
                        statistical_analysis=True,
                        splt_analysis=0,
                        n_meters=0,
                        performance_report=0)
    list(streamer)


from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, BatchNormalization, Conv1D, ReLU, MaxPooling1D, Flatten, Bidirectional, GRU
from tensorflow.keras.constraints import max_norm 
from . import CustomDistiller



class Distiller(CustomDistiller):
    def __init__(self, payload_size=784, header_fields_packets=32, n_classes=[]) -> None:
        super(Distiller, self).__init__(
            modalities=[
                wang_payload_modality(payload_size),
                lopez_protocol_header_fields_modality(header_fields_packets)
            ],
            adapter_size=128, 
            n_classes=n_classes
        )



def wang_payload_modality(payload_size=784):
    input_layer_payload_modality = Input(shape=(payload_size,1), name='input_payload')
    return Model(
        name='Wang payload modality (nbytes)',
        inputs=input_layer_payload_modality,
        outputs=stack([
            input_layer_payload_modality,
            Conv1D(16, 25, name='Conv1D_payload_1'),
            ReLU(),
            MaxPooling1D(3, name='MaxPooling1D_payload_1'),
            Conv1D(32, 35, name='Conv1D_payload_2'),
            ReLU(),
            MaxPooling1D(3, name='MaxPooling1D_payload_2'),
            Flatten(), 
        ])
    )


def lopez_protocol_header_fields_modality(packet_count=32):
    input_layer_protocol_fields_modality = Input(shape=(packet_count,4), name='input_protocol_fields')
    return Model(
        name='Lopez protocol header fields modality',
        inputs=input_layer_protocol_fields_modality,
        outputs=stack([
            input_layer_protocol_fields_modality,
            Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
            ReLU(),
            Flatten(),
        ])
    )



def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack



from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, BatchNormalization, Conv1D, ReLU, MaxPooling1D, Flatten, Bidirectional, GRU, LeakyReLU, Lambda, Dense, MaxPool2D, Conv2D, LSTM
from tensorflow.keras.constraints import max_norm 
from tensorflow import expand_dims
from . import CustomDistiller




class MalDist(CustomDistiller):
    def __init__(self, payload_size=784, header_fields_packets=32, n_classes=[]) -> None:
        super(MalDist, self).__init__(  
            modalities=[
                wang_payload_modality(payload_size),
                lopez_protocol_header_fields_modality(header_fields_packets),
                stnn_inspired_image_modality()
            ],
            adapter_size=128, 
            n_classes=n_classes
        )



def wang_payload_modality(payload_size=784):
    input_layer_payload_modality = Input(shape=(payload_size,1), name='input_payload')
    return Model(
        name='Wang payload modality (nbytes)',
        inputs=input_layer_payload_modality,
        outputs=stack([
            input_layer_payload_modality,
            Conv1D(16, 25, name='Conv1D_payload_1'),
            ReLU(),
            MaxPooling1D(3, name='MaxPooling1D_payload_1'),
            Conv1D(32, 35, name='Conv1D_payload_2'),
            ReLU(),
            MaxPooling1D(3, name='MaxPooling1D_payload_2'),
            Flatten(), 
        ])
    )


def lopez_protocol_header_fields_modality(packet_count=32):
    input_layer_protocol_fields_modality = Input(shape=(packet_count,4), name='input_protocol_fields')
    return Model(
        name='Lopez protocol header fields modality',
        inputs=input_layer_protocol_fields_modality,
        outputs=stack([
            input_layer_protocol_fields_modality,
            Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
            ReLU(),
            Flatten(),
        ])
    )


def stnn_inspired_image_modality():
    input_layer_stnn_modality = Input(shape=(5,14), name='input_stnn')
    return Model(
        name='STNN-inspired image modality',
        inputs=input_layer_stnn_modality,
        outputs=stack([
            input_layer_stnn_modality,
            Bidirectional(LSTM(65,return_sequences=True)),
            Lambda(lambda x: expand_dims(x, axis=3)),
            Conv2D(32,3,padding='same'),
            LeakyReLU(),
            Conv2D(32,3,padding='same'),
            LeakyReLU(),
            MaxPool2D(2),
            Conv2D(64,3,padding='same'),
            LeakyReLU(),
            Conv2D(128,3,padding='same'),
            LeakyReLU(),
            MaxPool2D(2),
            Flatten(),
            Dense(512),
        ])
    )

def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack



from models import DeepMALRawFlows
from temp.n_bytes_per_packet import NBytesPerPacket
from temp.nbytes_plugin import NBytes
from extractor import Extractor
import glob
from models import M1CNN
from os import path, mkdir, stat
import pandas as pd
import numpy as np
import ast
from sklearn.preprocessing import OneHotEncoder



def M1CNNPipeline():
    # 1. Declare PCAP files for feature extraction. - Labels should be on dirs - or filenames
    pcaps_dir_path='./temp/pcap_files_pipeline/'
    csv_storage_location = 'temp'
    features, labels = get_m1cnn_features(csv_storage_location, directory_based_labelling(dir_path=pcaps_dir_path)) 
    # 4. Declare/Create models.
    model = M1CNN(n_classes=len(labels[0]))
    # 5. Train models.
    model.fit(features, labels, epochs=3)

    

def DeepMALPipeline():
    # 1. Declare PCAP files for feature extraction. - Labels should be on dirs - or filenames
    pcaps_dir_path='./temp/pcap_files_pipeline/'
    csv_storage_location = 'temp'
    features, labels = get_deepmal_features(csv_storage_location, directory_based_labelling(dir_path=pcaps_dir_path)) 
    # 4. Declare/Create models.
    model = DeepMALRawFlows(n_classes=len(labels[0]))
    # 5. Train models.
    model.fit(features, labels, epochs=3)


def get_m1cnn_features(csv_storage_location, labelling_method):
    '''
    M1CNN
    '''
    custom_plugin_package = [NBytes()]
    make_dir_if_doesnt_exist(path.join(csv_storage_location, 'm1cnn'))
    extract_features_dir(path.join(csv_storage_location, 'm1cnn'), labelling_method, custom_plugin_package)
    # 3. Read/Load those pcap files to memory 
    return read_all_csvs_in_dir(csv_storage_location)
    

def get_deepmal_features(csv_storage_location, labelling_method):
    '''
    DeepMAL
    '''
    custom_plugin_package = [NBytesPerPacket(100, max_packets=2)]
    make_dir_if_doesnt_exist(path.join(csv_storage_location, 'deepmal'))
    extract_features_dir(path.join(csv_storage_location, 'deepmal'), labelling_method, custom_plugin_package)
    # 3. Read/Load those pcap files to memory 
    return read_all_csvs_in_dir_deepmal(csv_storage_location)


def extract_features_dir(csv_storage_location, labelling_method, custom_plugin_package):
    '''
    General Purpose
    '''
    for pcap_file, label in labelling_method:
        print(pcap_file, label)
        # 2. Extract features for each PCAP file.
        extract_features_pcap_file(pcap_file, csv_storage_location, label, custom_plugin_package)

    
def extract_features_pcap_file(pcap_file, csv_storage_location, label, custom_plugin_package):
    '''
    General Purpose
    '''
    output_dirpath = path.join(csv_storage_location, get_last_dir_in_path(pcap_file))
    if not make_dir_if_doesnt_exist(output_dirpath):
        if compare_last_modified(pcap_file, output_dirpath) == 1:
            print('Skipping feature extraction as CSVs already exist and up-to-date.')
            return
    sessions_csv_filepath = Extractor(pcap_file, output_dirpath, 'tcp or udp', custom_plugin_package).extract()
    # 2.1 add label column
    # Cache? 
    df = pd.read_csv(sessions_csv_filepath)
    df['label'] = np.full(len(df.index), label)
    df.to_csv(sessions_csv_filepath, index=False)
    
    
def read_all_csvs_in_dir(csv_storage_location):
    '''
    M1CNN
    '''
    dfs = []
    for subdir in glob.iglob(path.join(csv_storage_location, '*')):
        for csv_file in glob.iglob(path.join(subdir, '*sessions.csv')):
            dfs.append(pd.read_csv(csv_file, usecols=['udps.n_bytes', 'label']))
    assert (len(dfs) > 0), 'No CSV files were read.'
    df = pd.concat(dfs)
    df['udps.n_bytes'] = df['udps.n_bytes'].apply(ast.literal_eval)
    features = np.stack(df['udps.n_bytes'].values)
    enc = OneHotEncoder(handle_unknown='ignore')
    labels = np.stack(enc.fit_transform(df['label'].values.reshape(-1,1)).toarray())
    
    return features, labels
    
    
def read_all_csvs_in_dir_deepmal(csv_storage_location):
    '''
    deepmal
    '''
    dfs = []
    for subdir in glob.iglob(path.join(csv_storage_location, '*')):
        for csv_file in glob.iglob(path.join(subdir, '*sessions.csv')):
            dfs.append(pd.read_csv(csv_file, usecols=['udps.n_bytes_per_packet', 'label']))
    assert (len(dfs) > 0), 'No CSV files were read.'
    df = pd.concat(dfs)
    df['udps.n_bytes_per_packet'] = df['udps.n_bytes_per_packet'].apply(ast.literal_eval)
    df['udps.n_bytes_per_packet'] = df['udps.n_bytes_per_packet'].apply(np.array, dtype='float32')
    features = np.stack(df['udps.n_bytes_per_packet'].values)
    enc = OneHotEncoder(handle_unknown='ignore')
    labels = np.stack(enc.fit_transform(df['label'].values.reshape(-1,1)).toarray())
    
    return features, labels

    
def directory_based_labelling(dir_path):
    '''
    General Purpose
    '''
    for dir_with_label in glob.iglob(path.join(dir_path, '*')):
        label = get_last_dir_in_path(dir_with_label)
        for pcap_filename in glob.iglob(path.join(dir_with_label, '*.pcap'), recursive=True):
            yield pcap_filename, label


def filename_based_labelling(dir_path, labeller):
    '''
    General Purpose
    '''
    for pcap_filename in glob.iglob(path.join(dir_path, '**', '*.pcap'), recursive=True):
        label = labeller(get_last_dir_in_path(pcap_filename))
        yield pcap_filename, label
            
            
def compare_last_modified(path1, path2):
    last_modified_path1 = stat(path1).st_mtime
    last_modified_path2 = stat(path2).st_mtime
    if last_modified_path1 > last_modified_path2:
        return 0
    else:
        return 1


def make_dir_if_doesnt_exist(dirpath):
    '''
    General Purpose
    '''
    if not path.exists(dirpath):
        try:
            mkdir(dirpath)
            return True
        except:
            print('Error: failed to make a directory with path of', dirpath)
            return False
    return False

def get_last_dir_in_path(dir_path):
    '''
    General Purpose
    '''
    return path.basename(path.normpath(dir_path))
    

#M1CNNPipeline()
DeepMALPipeline()


from nfstream import NFPlugin
from hashlib import md5
from main import some_val

class HostLayerFeatures(NFPlugin):
    '''
        EXPERIMENTAL
    
    '''
    def __init__(self, **kwargs): 
        super().__init__(**kwargs)
        self.hosts = dict()

    def on_init(self, packet, flow):
        id = self._five_tuple_string_of(flow)
        if flow.src_ip not in self.hosts:
            self.hosts[flow.src_ip] = Host(flow.src_ip)
        if flow.dst_ip not in self.hosts:
            self.hosts[flow.dst_ip] = Host(flow.dst_ip)
        

    def on_update(self, packet, flow):
        src_host = self.hosts[packet.src_ip]
        src_host.sent_packets += 1
        
        global some_val
        some_val += 1
        
        

    def _five_tuple_string_of(self, obj):
        return md5(
            str.join('',
                sorted([
                    obj.src_ip,
                    str(obj.src_port),
                    obj.dst_ip,
                    str(obj.dst_port),
                    str(obj.protocol)])))

        
        
        
class Host:
    def __init__(self, ip: str) -> None:
        self.ip = ip
        # default values
        self.recv_packets = 0
        self.recv_bytes   = 0
        self.sent_packets = 0
        self.sent_bytes   = 0
    
    
    def __hash__(self) -> int:
        pass

    def __str__(self) -> str:
        '''
        for debug purposes.
        '''
        
        pass


from nfstream import NFPlugin
from scapy.all import IP, TLS, load_layer

load_layer("tls")
# requires scapy 2.4+

class TLSRecord(NFPlugin):
    '''
    WORK IN PROGRESS!!
        Extracts the following features:
        - Number of TLS packets.
        - Number of TLS records.
        - TLS record size.
        - TLS record type: change_cipher_spec(20), alert(21), handshake(22), application_data(23).
        - TLS record direction (src2dst = 0, dst2src = 1).

        Note that due to TCP segmentation, the plugin can sometimes
        not detect TLS records as they might be segmented across
        multiple packets, this plugin does not try to reassmble TCP stream
        in order to recover missing TLS records.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''

        flow.udps.src2dst_tls_packets = 0 # number of tls packets
        flow.udps.dst2src_tls_packets = 0
        flow.udps.bidirectional_tls_packets = 0
        flow.udps.src2dst_tls_records = 0
        flow.udps.dst2src_tls_records = 0
        flow.udps.bidirectional_tls_records = 0
        flow.udps.src2dst_tls_record_size_freq = list()
        flow.udps.dst2src_tls_record_size_freq = list()
        flow.udps.bidirectional_tls_record_size_freq = list()
        flow.udps.src2dst_tls_record_type_freq = {'handshake':0, 'application':0} # limited to only handshake and application types.
        flow.udps.dst2src_tls_record_type_freq = {'handshake':0, 'application':0}
        flow.udps.bidirectional_tls_record_type_freq = {'handshake':0, 'application':0} 

        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        ip_packet = IP(packet.ip_packet)
        if TLS in ip_packet:
            if packet.direction == 0: # src2dst
                flow.udps.src2dst_tls_packets += 1
                #self._increment_dict_value(flow.udps.src2dst_tls_record_type_freq, ?)

    def on_expire(self, flow):
        pass

    def _increment_dict_value(self, d :dict, key: int):
        if key not in d:
            d[key] = 0
        d[key] += 1




'''
Note: 
    Requires python 3
    Requires linux (can use WSL)
    Doesn't work with pcapng suffix files, only pcap suffixes 
        work around is just rename the suffix from pcapng to pcap.

Using Scapy library as a complement for NFStream such as parsing DNS packets
'''

from plugins.asn_info import ASNInfo
from sessions_processor import SessionsProcessor
from plugins.n_pkts_byte_freq import NPacketsByteFrequency
from plugins.first_packet_payload import FirstPacketPayloadLen
from plugins.most_freq_payload_len_ratio import MostFreqPayloadLenRatio
from plugins.dns_counter import DNSCounter
from plugins.small_pkt_payload_ratio import SmallPacketPayloadRatio
from plugins.pkt_rel_time import PacketRelativeTime
from plugins.res_req_diff_time import ResReqDiffTime
from plugins.byte_freq import ByteFrequency
from plugins.graypic import GrayPic1
from flows_processor import FlowsProcessor
from nfstream import NFStreamer  # https://www.nfstream.org/docs/api

# possible files:
# './pcaps/tls.pcap' 
# "./pcaps/DoH-Firefox84-NextDNS-1.pcap"
# "./pcaps/merged.pcap"
# "./tests/pcaps/dns_1.pcap"
# "./pcaps/DoH-Firefox84-first-100-sec.pcap"
pcap_filepath = "./pcaps/DoH-Firefox84-NextDNS-1.pcap"
bpf_filter_string = None
plugins = [ASNInfo(),
           DNSCounter(),
           FirstPacketPayloadLen(),
           MostFreqPayloadLenRatio(),
           NPacketsByteFrequency(n_first_packets=6),
           PacketRelativeTime(),
           SmallPacketPayloadRatio(),
           ResReqDiffTime()]
           
my_streamer = NFStreamer(source=pcap_filepath,
                         decode_tunnels=True,
                         bpf_filter=bpf_filter_string,
                         promiscuous_mode=True,
                         snapshot_length=1536,
                         idle_timeout=1,
                         active_timeout=1,
                         accounting_mode=3,
                         udps=plugins,
                         n_dissections=20,
                         statistical_analysis=True,
                         splt_analysis=0,
                         n_meters=0,
                         performance_report=0)

my_streamer.to_csv('out-timed-flows.csv', columns_to_anonymize=[])

session_streamer = NFStreamer(source=pcap_filepath,
                         decode_tunnels=True,
                         bpf_filter=bpf_filter_string,
                         promiscuous_mode=True,
                         snapshot_length=1536,
                         idle_timeout=99999999,
                         active_timeout=99999999,
                         accounting_mode=3,
                         udps=plugins,
                         n_dissections=20,
                         statistical_analysis=True,
                         splt_analysis=0,
                         n_meters=0,
                         performance_report=0)

session_streamer.to_csv('out-sessions.csv', columns_to_anonymize=[])

'''
df = my_streamer.to_pandas(columns_to_anonymize=[])
sp = SessionsProcessor()
sp.process(df)


fp = FlowsProcessor(my_streamer)
fp.process()
'''



import ast
import pandas as pd
import os
from enum import IntEnum
import numpy as np
from pandas.core.groupby.generic import DataFrameGroupBy

class Protocol(IntEnum):
    TCP=6
    UDP=17


class HostsProcessor:
    '''
        Works on pre-computed pandas dataframe of bidirectional
        network flows of 5 tuple.
    '''
    def __init__(self, df, output_dirpath) -> None:
        self.df = df
        self.output_dirpath = output_dirpath
        self.unidir_grouped_by_source = df.groupby(by='src_ip')
        self.unidir_grouped_by_dest = df.groupby(by='dst_ip')

    def process(self):
        '''
        Note: code here can be optimized both in terms of performance and memory consumption.
        But at the moment, readability is preferable.
        '''
        # output dataframe
        out_df = pd.DataFrame()
        # pre proccessing to df
        self.df['udps.src2dst_payload_freq'] = self.df['udps.src2dst_payload_freq'].apply(ast.literal_eval)
        self.df['udps.dst2src_payload_freq'] = self.df['udps.dst2src_payload_freq'].apply(ast.literal_eval)
        # Features
        # first seen (ms)
        first_seen_ms = self._host(self.df, 'bidirectional_first_seen_ms', 'bidirectional_first_seen_ms').min(numeric_only=True).squeeze()
        out_df['first_seen_ms'] = first_seen_ms
        # last seen (ms)
        last_seen_ms = self._host(self.df, 'bidirectional_last_seen_ms', 'bidirectional_last_seen_ms').min(numeric_only=True).squeeze()
        out_df['last_seen_ms'] = last_seen_ms
        # seen duration (ms)
        seen_duration_ms = last_seen_ms - first_seen_ms
        out_df['seen_duration_ms'] = seen_duration_ms
        # packets sent by host
        sent_packets_df = self._host(self.df, 'src2dst_packets', 'dst2src_packets')
        sent_packets = sent_packets_df.sum().squeeze()
        out_df['sent_packets'] = sent_packets
        # packets received by host
        received_packets = self._host(self.df, 'dst2src_packets', 'src2dst_packets').sum().squeeze()
        out_df['received_packets'] = received_packets
        # received/sent packets ratio by host
        recv_sent_packets_ratio = received_packets / sent_packets
        out_df['recv_sent_packets_ratio'] = recv_sent_packets_ratio
        # average of packets sent in sessions by host
        sent_mean_packets_per_sessions = sent_packets_df.mean().squeeze()
        out_df['sent_mean_packets_per_sessions'] = sent_mean_packets_per_sessions
        # median of packets sent in sessions by host
        sent_median_packets_per_sessions = sent_packets_df.median().squeeze()
        out_df['sent_median_packets_per_sessions'] = sent_median_packets_per_sessions
        # std deviation of packets sent in sessions by host
        sent_stdev_packets_per_sessions = sent_packets_df.std().squeeze()
        out_df['sent_stdev_packets_per_sessions'] = sent_stdev_packets_per_sessions
        # std deviation of packets sent in sessions by host
        sent_var_packets_per_sessions = sent_packets_df.var().squeeze()
        out_df['sent_variance_packets_per_sessions'] = sent_var_packets_per_sessions
        # bytes sent
        sent_bytes = self._host(self.df, 'src2dst_bytes', 'dst2src_bytes').sum().squeeze()
        out_df['sent_bytes'] = sent_bytes
        # bytes received
        recv_bytes = self._host(self.df, 'dst2src_bytes', 'src2dst_bytes').sum().squeeze()
        out_df['recv_bytes'] = recv_bytes
        # bytes received
        recv_sent_bytes_ratio = recv_bytes / sent_bytes
        out_df['recv_sent_bytes_ratio'] = recv_sent_bytes_ratio
        # sent payload frequency by host
        sent_payload_freq = self._host(self.df, 'udps.src2dst_payload_freq', 'udps.dst2src_payload_freq').apply(
            lambda row: self._merge_dicts(row['udps.src2dst_payload_freq'])
            )
        out_df['sent_payload_freq'] = sent_payload_freq
        # most frequent payload length sent by host
        def dict_max(dict1):
            return max(dict1.values())
        sent_most_freq_payload_len = sent_payload_freq.apply(dict_max)
        out_df['sent_most_freq_payload_len'] = sent_most_freq_payload_len
        # most freq payload sent to number of packets sent ratio by host
        sent_most_freq_payload_ratio = sent_most_freq_payload_len / sent_packets
        out_df['sent_most_freq_payload_ratio'] = sent_most_freq_payload_ratio
        # number of distinct different destination addresses
        unique_dst_addresses = self._host(self.df, 'dst_ip', 'src_ip').nunique().squeeze()
        out_df['num_distinct_dst_addresses'] = unique_dst_addresses
        # number of distinct different destination ports
        unique_dst_ports = self._host(self.df, 'dst_port', 'src_port').nunique().squeeze()
        out_df['num_distinct_dst_ports'] = unique_dst_ports
        # number of small packets sent
        sent_small_pkt_payload_packets = self._host(self.df, 'udps.src2dst_small_packet_payload_packets', 
            'udps.dst2src_small_packet_payload_packets').sum().squeeze()
        out_df['sent_small_packet_payload_packets'] = sent_small_pkt_payload_packets
        # number of small packets to the number of sent packets ratio
        sent_small_pkt_payload_ratio = sent_small_pkt_payload_packets / sent_packets
        out_df['sent_small_packet_payload_ratio'] = sent_small_pkt_payload_ratio
        # number of total sessions
        sessions = self._host(self.df, 'protocol', 'protocol').count().squeeze()
        out_df['sessions'] = sessions
        # number of UDP sessions
        udp_sessions = self._host(self.df[self.df['protocol']==Protocol.UDP], 'protocol', 'protocol').count().squeeze()
        out_df['udp_sessions'] = udp_sessions
        # number of TCP sessions
        tcp_sessions = self._host(self.df[self.df['protocol']==Protocol.TCP], 'protocol', 'protocol').count().squeeze()
        out_df['tcp_sessions'] = tcp_sessions
        # number of UDP packets sent
        sent_udp_packets = self._host(self.df[self.df['protocol']==Protocol.UDP], 'src2dst_packets', 'dst2src_packets').sum().squeeze()
        out_df['sent_udp_packets'] = sent_udp_packets
        # number of TCP packets sent
        sent_tcp_packets = self._host(self.df[self.df['protocol']==Protocol.TCP], 'src2dst_packets', 'dst2src_packets').sum().squeeze()
        out_df['sent_tcp_packets'] = sent_tcp_packets
        # number of DNS packets sent
        sent_dns_packets = self._host(self.df[self.df['protocol']==Protocol.UDP], 
                                      'udps.src2dst_dns_packets', 
                                      'udps.dst2src_dns_packets').sum().squeeze()
        out_df['sent_dns_packets'] = sent_dns_packets
        # ratio of number UDP packets to number of packets sent
        sent_udp_packets_ratio = sent_udp_packets / sent_packets
        out_df['sent_udp_packets_ratio'] = sent_udp_packets_ratio
        # ratio of number TCP packets to number of packets sent
        sent_tcp_packets_ratio = sent_tcp_packets / sent_packets
        out_df['sent_tcp_packets_ratio'] = sent_tcp_packets_ratio
        # ratio of number DNS packets to number of packets sent
        sent_dns_packets_ratio = sent_dns_packets / sent_packets
        out_df['sent_dns_packets_ratio'] = sent_dns_packets_ratio

        

        #save to file
        out_df.to_csv(os.path.join(self.output_dirpath, 'out-hosts.csv'))
        
    def _host(self, df, k1, k2):
        '''
        Description:
            Retrieves all relavent columns from both direction that represent the same thing.
            In case we want all packets sent by some source IP address (host), the function
            combines the packets sent in session when the host is the initiator of the session,
            with the number of packets received in the session where the host is not the initator.
            This function do this for all hosts in the df in seperate.
            
        Params:
            df: the sessions dataframe created by NFStream.
            k1: the key/column for the df to extract its values from.
            k2: the equivalent key/column for k1 just for the reverse direction.
            
        Example:
            If the df have columns 'src2dst_packets' and 'dst2src_packets' that represents
            the number of packets from source to destination and from destination to source respectfuly,
            then `k1='src2dst_packets'` and `k2='dst2src_packets'`
            so the function call would be:
            `self._host(df, 'src2dst_packets', 'dst2src_packets')`
        '''
        unidir_grouped_by_source = self.unidir_grouped_by_source[['src_ip', k1]]
        unidir_grouped_by_dest   = self.unidir_grouped_by_dest[['dst_ip', k2]]
        dataframes = list()
        for group_name, group in unidir_grouped_by_source:
            if group_name in unidir_grouped_by_dest.groups:
                dest_group = unidir_grouped_by_dest.get_group(group_name)
                dest_group = dest_group.rename(columns={'dst_ip':'src_ip',k2:k1})
                merged_group = pd.concat([group[['src_ip', k1]], dest_group])
                dataframes.append(merged_group)
            else:
                dataframes.append(group[['src_ip', k1]])

        return pd.concat(dataframes).groupby(by='src_ip', as_index=True)
    
    def _stats(self, grouped_df: DataFrameGroupBy):
        '''
        Description:
            Computes a pre-defined list of statistical functions
            for a grouped dataframe (DataFrameGroupBy) object.
            The statistical functions are:
            - mean
            - median
            - std
            - var
            in that order.
            
        Example:
            out_df[['mean', 'median', 'std', 'var']] = self._stats(grouped_df)
        '''
        return grouped_df.agg([
            np.mean,
            np.median,
            np.std,
            np.var
        ])
    
    def _merge_dicts(self, dicts):
        ''' Merge dictionaries and keep values of common keys in list'''
        merged_dict = dict()
        for i in range(len(dicts)):
            merged_dict = self._merge_dict(merged_dict, dicts.iloc[i])
        
        return merged_dict

    def _merge_dict(self, dict1, dict2):
        ''' Merge dictionaries and keep values of common keys in list'''
        dict3 = {**dict1, **dict2}
        for key, value in dict3.items():
            if key in dict1 and key in dict2:
                    dict3[key] = value + dict1[key]
        return dict3




if __name__ == '__main__':
    print('starting')
    df = pd.read_csv('./temp/out-sessions.csv')
    HostsProcessor(df ,'temp').process()


from runstats import *
from nfstream import NFPlugin
import numpy as np
import statistics


class Packets_size_and_interarrival_time(NFPlugin):
    '''
    W.I.P:
        1. Optimize median calculations by using histogram (if possible)
        2. Add info about this plugin and its origin paper.
    '''
    def on_init(self, packet, flow):
        flow.udps.packets_size = list()
        flow.udps.packets_interarrival_time = list()
        self.on_update(packet,flow)

    def on_update(self, packet, flow):
        flow.udps.packets_size.append(packet.raw_size)
        flow.udps.packets_interarrival_time.append(packet.delta_time)

    def on_expire(self, flow):
        packets_size_statistics = Statistics(flow.udps.packets_size)
        packets_interarrival_time_statistics = Statistics(flow.udps.packets_interarrival_time)

        #Packets Size Statistical
        packets_size_Q1,packets_size_Q2,packets_size_Q3 = quartile(np.array(flow.udps.packets_size))
        flow.udps.packets_size_min = packets_size_statistics.minimum() if packets_size_statistics._count > 0 else 0
        flow.udps.packets_size_max = packets_size_statistics.maximum() if packets_size_statistics._count > 0 else 0
        flow.udps.packets_size_stddev = packets_size_statistics.stddev() if packets_size_statistics._count >= 2 else 0
        flow.udps.packets_size_first_quartile = packets_size_Q1
        flow.udps.packets_size_second_quartile = packets_size_Q2
        flow.udps.packets_size_third_quartile = packets_size_Q3
        flow.udps.packets_size_mean = packets_size_statistics.mean()
        flow.udps.packets_size_median_absoulte_deviation = median_absolute_deviation(np.array(flow.udps.packets_size))
        flow.udps.packets_size_variance = packets_size_statistics.variance() if packets_size_statistics._count >= 2 else 0
        flow.udps.packets_size_skewness = packets_size_statistics.skewness() if flow.udps.packets_size_stddev != 0 else 0
        flow.udps.packets_size_kurtosis = packets_size_statistics.kurtosis() if flow.udps.packets_size_stddev != 0 else 0
        flow.udps.packets_size_sum = sum(flow.udps.packets_size)
        
        #Packets Interarrival Time Statistical
        packets_interarrival_time_Q1,packets_interarrival_time_Q2,packets_interarrival_time_Q3 = quartile(np.array(flow.udps.packets_interarrival_time))
        flow.udps.packets_interarrival_time_min = packets_interarrival_time_statistics.minimum() if packets_interarrival_time_statistics._count > 0 else 0
        flow.udps.packets_interarrival_time_max = packets_interarrival_time_statistics.maximum() if packets_interarrival_time_statistics._count > 0 else 0
        flow.udps.packets_interarrival_time_stddev = packets_interarrival_time_statistics.stddev() if packets_interarrival_time_statistics._count >= 2 else 0
        flow.udps.packets_interarrival_time_first_quartile = packets_interarrival_time_Q1
        flow.udps.packets_interarrival_time_second_quartile = packets_interarrival_time_Q2
        flow.udps.packets_interarrival_time_third_quartile = packets_interarrival_time_Q3
        flow.udps.packets_interarrival_time_mean = packets_interarrival_time_statistics.mean()
        flow.udps.packets_interarrival_time_median_absoulte_deviation = median_absolute_deviation(np.array(flow.udps.packets_interarrival_time))
        flow.udps.packets_interarrival_time_variance = packets_interarrival_time_statistics.variance() if packets_interarrival_time_statistics._count >= 2 else 0
        flow.udps.packets_interarrival_time_skewness = packets_interarrival_time_statistics.skewness() if flow.udps.packets_interarrival_time_stddev != 0 else 0
        flow.udps.packets_interarrival_time_kurtosis = packets_interarrival_time_statistics.kurtosis() if flow.udps.packets_interarrival_time_stddev != 0 else 0
        flow.udps.packets_interarrival_time_sum = sum(flow.udps.packets_interarrival_time)

def median_absolute_deviation(arr):
    return np.median(np.absolute(arr-np.median(arr)))

def quartile(arr):
    sorted_arr = np.sort(arr)
    median = sorted_median(sorted_arr)
    sorted_len = len(sorted_arr)
    half = int(sorted_len/2)
    if sorted_len % 2 == 0:
        return sorted_median(sorted_arr[0:half]),median,sorted_median(sorted_arr[half:sorted_len])
    else:
        return sorted_median(sorted_arr[0:half+1]),median,sorted_median(sorted_arr[half:sorted_len])

def sorted_median(sorted_arr):
    arr_len = len(sorted_arr)
    half = int(arr_len/2)
    if arr_len % 2 == 0:
        return (sorted_arr[half-1]+sorted_arr[half])/2
    else:
        return sorted_arr[half]


from nfstream import NFStreamer, NFPlugin
import pandas as pd
import numpy as np
from scapy.all import IP, DNS, IPv6, TCP, rdpcap, load_layer, raw
load_layer("tls")


class TLSPlugin(NFPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        flow.udps.ack = 0
        flow.udps.seq = 0
        flow.udps.record_len_up = 0
        flow.udps.record_len_down = 0
        flow.udps.record_lens = []
        
        
        self.on_update(packet, flow)
    
    def on_update(self, packet, flow):
        if flow.protocol == 6 and (flow.src_port == 443 or flow.dst_port == 443) and packet.payload_size > 0:
            if packet.direction == 0:
                record_lens, leftover = self.process(packet, flow.udps.record_len_up)
                flow.udps.record_len_up = leftover
                flow.udps.record_lens.extend(record_lens)
            else:
                record_lens, leftover = self.process(packet, flow.udps.record_len_down)
                flow.udps.record_len_down = leftover
                flow.udps.record_lens.extend(record_lens)
    
    def on_expire(self, flow):
        pass

    def process(self, packet, record_len, tls_header_len=5):
        raw_tcp_payload_data = raw(IP(packet.ip_packet)[TCP].payload)
        record_lens = []
        pos = 0
        while pos < len(raw_tcp_payload_data):
            if record_len == 0: # empty/new
                record_len = int.from_bytes(raw_tcp_payload_data[pos+3:pos+5], byteorder='big')
                record_lens.append(record_len)
                pos += tls_header_len
                
            received_data_len = len(raw_tcp_payload_data[pos:])
            if record_len <= received_data_len:
                pos += record_len
                received_data_len = 0
                record_len = 0
            else: 
                record_len -= received_data_len
                pos += received_data_len
        
        return (record_lens, record_len)



source = "/mnt/d/temp/University/masters/thesis/research/temp/NFStream/pcaps/DoH-Chrome81-Cloudflare.pcap"  #temp/skype_video2b.pcap' #'temp/DoH-Firefox84-NextDNS-1-pcap-format.pcap' # 'temp/skype_video2b.pcap' # sys.argv[1]


def entry():
    streamer = NFStreamer(source=source,
                                    decode_tunnels=True,
                                    bpf_filter=None,
                                    promiscuous_mode=True,
                                    snapshot_length=1536,
                                    idle_timeout=99999999,
                                    active_timeout=99999999,
                                    accounting_mode=3,
                                    udps=[
                                        #FlowPic2019('./temp', 0,  flow_active_time=60)
                                        #FlowPic2('./temp', time_per_subflow=60)
                                        #GrayPic1('./temp')
                                        TLSPlugin()
                                    ],
                                    n_dissections=20,
                                    statistical_analysis=True,)
    streamer.to_pandas()

if __name__ == '__main__':
    entry()


import os

from nfstream.streamer import NFStreamer

from plugins.flow_time_slicer import FlowTimeSlicer
from plugins.most_freq_payload_len_ratio import MostFreqPayloadLenRatio
from plugins.is_dns import IsDNS
from plugins.small_pkt_payload_ratio import SmallPacketPayloadRatio
from plugins.pkt_rel_time import PacketRelativeTime
from plugins.res_req_diff_time import ResReqDiffTime
from plugins.byte_freq import ByteFrequency
from flows_processor import FlowsProcessor

class PCAPPreprocessor:
    '''
        Windows, for dev-testing.
    '''
    def __init__(self, pcap_filepath) -> None:
        self.pcap_filepath = pcap_filepath
    
    def preprocess(self): 
        self._split_to_sessions(self.pcap_filepath)
        dir_list = os.listdir('./pcaps/splitted_pcap/')
        self._create_editcap_directory()
        for filepath in dir_list:
            self._split_to_time_windows(filepath)
    
    
    def _split_to_sessions(self, pcap_filepath):
        cd_command = 'cd ./pcaps'
        command = '"SplitCap.exe" -r ' + pcap_filepath + ' -o splitted_pcap -s sessions -d'
        op_result = os.system(cd_command + ' && ' + command)
        
        if op_result != 0:
            raise 'Failed to split PCAP file to sessions. command: ' + command
        
    def _split_to_time_windows(self, pcap_filepath, seconds = 10):
        filepath = './pcaps/splitted_pcap/' + pcap_filepath
        timesplit_command = '"D:/Program Files/Wireshark/editcap.exe" -i ' + str(seconds) + ' ' + filepath + ' ./pcaps/editcap_output/file.pcap'
        print(timesplit_command)
        op_result = os.system(timesplit_command)
        
        if op_result != 0:
            raise 'Failed to split PCAP file to time windows. command: ' + timesplit_command
    
    def _create_editcap_directory(self):
        path = './pcaps/editcap_output'
        try:
            os.mkdir(path)
        except OSError:
            print ("Creation of the directory %s failed" % path)
        else:
            print ("Successfully created the directory %s " % path)
            
            
            
## for debug
'''
pcappp = PCAPPreprocessor('DoH-Firefox84-NextDNS-1-pcap-format.pcap')
pcappp.preprocess()
'''


dir_list = os.listdir('./pcaps/editcap_output/')
for pcap_filepath_windowed in dir_list:
    pcap_filepath_windowed = './pcaps/editcap_output/' + pcap_filepath_windowed
    bpf_filter_string = None
    plugins = [FlowTimeSlicer(time=15),
            ResReqDiffTime(),
            ByteFrequency(),
            PacketRelativeTime(),
            SmallPacketPayloadRatio(),
            IsDNS(),
            MostFreqPayloadLenRatio()]
    streamer = NFStreamer(source=pcap_filepath_windowed,
                                decode_tunnels=True,
                                bpf_filter=bpf_filter_string,
                                promiscuous_mode=True,
                                snapshot_length=1536,
                                idle_timeout=1200,
                                active_timeout=1800,
                                accounting_mode=3,
                                udps=plugins,
                                n_dissections=20,
                                statistical_analysis=True,
                                splt_analysis=0,
                                n_meters=0,
                                performance_report=0)
    
    flows = list(streamer)
    print('Flow count:', len(flows))
print('Exiting...')



from math import e
from nfstream import NFStreamer, NFPlugin
import numpy as np
from scapy.all import IP, TCP
from runstats import *

class STNN(NFPlugin):
    '''
    wip
    '''
    def __init__(self,n_packets=32):
        self.n_packets = n_packets

    def on_init(self, packet, flow):
        
        flow.udps.bidirectional_addtional_info = np.full(4,0)
        flow.udps.bidirectional_packets_iat_stats = Statistics()
        flow.udps.bidirectional_packets_size_stats = Statistics()

        flow.udps.src2dst_addtional_info = np.full(4,0)
        flow.udps.src2dst_packets_iat_stats = Statistics()
        flow.udps.src2dst_packets_size_stats = Statistics()

        flow.udps.dst2src_addtional_info = np.full(4,0)
        flow.udps.dst2src_packets_iat_stats = Statistics()
        flow.udps.dst2src_packets_size_stats = Statistics()

        flow.udps.handshake_addtional_info = np.full(4,0)
        flow.udps.handshake_packets_iat_stats = Statistics()
        flow.udps.handshake_packets_size_stats = Statistics()
        flow.udps.handshake_packets_duration = 0

        flow.udps.data_addtional_info = np.full(4,0)
        flow.udps.data_packets_iat_stats = Statistics()
        flow.udps.data_packets_size_stats = Statistics()


        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if flow.bidirectional_packets <= self.n_packets:
            flow.udps.bidirectional_packets_iat_stats.push(packet.delta_time)
            flow.udps.bidirectional_packets_size_stats.push(packet.raw_size)

            if packet.direction == 0:
                flow.udps.src2dst_packets_iat_stats.push(packet.delta_time)
                flow.udps.src2dst_packets_size_stats.push(packet.raw_size)
            elif packet.direction == 1:
                flow.udps.dst2src_packets_iat_stats.push(packet.delta_time)
                flow.udps.dst2src_packets_size_stats.push(packet.raw_size)
            if packet.protocol == 6 and packet.syn == True:
                flow.udps.handshake_packets_iat_stats.push(packet.delta_time)
                flow.udps.handshake_packets_size_stats.push(packet.raw_size)
                flow.udps.handshake_addtional_info[0] += 1
                flow.udps.handshake_addtional_info[1] += packet.raw_size
                flow.udps.handshake_packets_duration = flow.bidirectional_duration_ms
            if packet.protocol != 6 or packet.syn == False:
                flow.udps.data_packets_iat_stats.push(packet.delta_time)
                flow.udps.data_packets_size_stats.push(packet.raw_size)
        
        



    def on_expire(self, flow):
        matrix = np.empty((5,14))
        bidirectional_duration = 1 if flow.bidirectional_duration_ms == 0 else flow.bidirectional_duration_ms/1000
        flow.udps.bidirectional_addtional_info = [flow.bidirectional_packets,flow.bidirectional_bytes,flow.bidirectional_packets/bidirectional_duration,flow.bidirectional_bytes/bidirectional_duration]
        src2dst_duration = 1 if flow.src2dst_duration_ms == 0 else flow.src2dst_duration_ms/1000
        flow.udps.src2dst_addtional_info = [flow.src2dst_packets,flow.src2dst_bytes,flow.src2dst_packets/src2dst_duration,flow.src2dst_bytes/src2dst_duration]
        dst2src_duration = 1 if flow.dst2src_duration_ms == 0 else flow.dst2src_duration_ms/1000
        flow.udps.dst2src_addtional_info = [flow.dst2src_packets,flow.dst2src_bytes,flow.dst2src_packets/dst2src_duration,flow.dst2src_bytes/dst2src_duration]
        handshake_duration = flow.udps.handshake_packets_duration
        flow.udps.handshake_packets_duration = 1 if flow.udps.handshake_packets_duration == 0 else flow.udps.handshake_packets_duration/1000
        flow.udps.handshake_addtional_info[2:] = [flow.udps.handshake_addtional_info[0]/flow.udps.handshake_packets_duration,flow.udps.handshake_addtional_info[1]/flow.udps.handshake_packets_duration]
        data_packets = flow.bidirectional_packets-flow.udps.handshake_addtional_info[0]
        data_bytes = flow.bidirectional_bytes-flow.udps.handshake_addtional_info[1]
        data_duration = 1 if flow.bidirectional_duration_ms-handshake_duration == 0 else (flow.bidirectional_duration_ms-handshake_duration)/1000
        flow.udps.data_addtional_info = [data_packets,data_bytes,data_packets/data_duration,data_bytes/data_duration]
        stnn_objects = [
            (flow.udps.bidirectional_addtional_info,flow.udps.bidirectional_packets_iat_stats,flow.udps.bidirectional_packets_size_stats),
            (flow.udps.src2dst_addtional_info,flow.udps.src2dst_packets_iat_stats,flow.udps.src2dst_packets_size_stats),
            (flow.udps.dst2src_addtional_info,flow.udps.dst2src_packets_iat_stats,flow.udps.dst2src_packets_size_stats),
            (flow.udps.handshake_addtional_info,flow.udps.handshake_packets_iat_stats,flow.udps.handshake_packets_size_stats),
            (flow.udps.data_addtional_info,flow.udps.data_packets_iat_stats,flow.udps.data_packets_size_stats),
        ]

        for i,(add_info,iat_stats,size_stats) in enumerate(stnn_objects):
            matrix[i,0] = iat_stats.minimum() if iat_stats._count > 0 else 0
            matrix[i,1] = iat_stats.maximum() if iat_stats._count > 0 else 0
            matrix[i,2] = iat_stats.mean()
            matrix[i,3] = iat_stats.stddev() if iat_stats._count >= 2 else 0
            matrix[i,4] = iat_stats.skewness() if matrix[i,3] != 0 else 0
            matrix[i,5] = size_stats.minimum() if size_stats._count > 0 else 0
            matrix[i,6] = size_stats.maximum() if size_stats._count > 0 else 0
            matrix[i,7] = size_stats.mean()
            matrix[i,8] = size_stats.stddev() if size_stats._count >= 2 else 0
            matrix[i,9] = size_stats.skewness() if matrix[i,8] != 0 else 0
            matrix[i,10:] = add_info
            
        del flow.udps.bidirectional_addtional_info
        del flow.udps.bidirectional_packets_iat_stats
        del flow.udps.bidirectional_packets_size_stats
        del flow.udps.src2dst_addtional_info
        del flow.udps.src2dst_packets_iat_stats
        del flow.udps.src2dst_packets_size_stats
        del flow.udps.dst2src_addtional_info
        del flow.udps.dst2src_packets_iat_stats
        del flow.udps.dst2src_packets_size_stats
        del flow.udps.handshake_addtional_info
        del flow.udps.handshake_packets_iat_stats
        del flow.udps.handshake_packets_size_stats
        del flow.udps.data_addtional_info
        del flow.udps.data_packets_iat_stats
        del flow.udps.data_packets_size_stats

        flow.udps.stnn_image = [list(i) for i in matrix.astype(np.float32)]


from nfstream import NFPlugin
from PIL import Image
import numpy as np
import os

class FlowPic2019(NFPlugin):
    '''
    File name structure:
        flowpic-[flow_start_time_ms]-[src_ip]-[src_port]-[dst_ip]-[dst_port]-[protocol]-[direction].png
    
    '''
    def __init__(self, save_path, flow_direction=0, flow_active_time=30, image_dims=(1500, 1500), **kwargs):
        super().__init__(**kwargs)
        self.save_path = save_path
        self.flowpic_flow_direction = flow_direction # NOT USED AT THE MOMENT
        self.flow_active_time = flow_active_time
        self.image_dims = image_dims

    def on_init(self, packet, flow):
        flow.udps.src2dst_flowpic_image = np.zeros(self.image_dims, dtype=np.uint32)
        flow.udps.dst2src_flowpic_image = np.zeros(self.image_dims, dtype=np.uint32)
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        if packet.ip_size == 0 or packet.ip_size > self.image_dims[0]:
            # discard
            return
        
        if packet.direction == 0: # src -> dst
            relative_time = packet.time \
                            - flow.src2dst_first_seen_ms
            flow.udps.src2dst_flowpic_image[packet.ip_size, self._scale_to(relative_time, 0, self.image_dims[1])] += 1
        else:
            relative_time = packet.time \
                            - flow.dst2src_first_seen_ms
            flow.udps.dst2src_flowpic_image[packet.ip_size, self._scale_to(relative_time, 0, self.image_dims[1])] += 1    
        
    def on_expire(self, flow):
        ## SRC -> DST
        # normalize all cells/pixels in the image to [0,255].
        #flow.udps.src2dst_flowpic_image = flow.udps.src2dst_flowpic_image * 255 / np.sum(flow.udps.src2dst_flowpic_image)
        # reverse the Y axis such that the higher you go, the higher the values.
        flow.udps.src2dst_flowpic_image = np.flip(flow.udps.src2dst_flowpic_image, axis=0) 
        
        # for clarity/debug:
        temp = np.copy(flow.udps.src2dst_flowpic_image)
        '''
        flow.udps.src2dst_flowpic_image[temp>0]  = 0
        flow.udps.src2dst_flowpic_image[temp==0] = 255
        flow.udps.src2dst_flowpic_image = flow.udps.src2dst_flowpic_image.astype(np.uint8)
        '''
        
        np.savez_compressed(os.path.join(
                    self.save_path,
                    '-'.join(['flowpic',str(flow.bidirectional_first_seen_ms), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol), 'src2dst'])
                ), 
                flowpic=flow.udps.src2dst_flowpic_image,)
        img = Image.fromarray(flow.udps.src2dst_flowpic_image , 'I') # L = 8bit black and white (gray scale)
        img.save(
            os.path.join(
                self.save_path,
                '-'.join(['flowpic',str(flow.bidirectional_first_seen_ms), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol), 'src2dst'])
                + '.png'))
        ## DST -> SRC
        # normalize all cells/pixels in the image to [0,255].
        #flow.udps.dst2src_flowpic_image = flow.udps.dst2src_flowpic_image * 255 / np.sum(flow.udps.dst2src_flowpic_image)
        # reverse the Y axis such that the higher you go, the higher the values.
        flow.udps.dst2src_flowpic_image = np.flip(flow.udps.dst2src_flowpic_image, axis=0) 
        
        # for clarity/debug:
        temp = np.copy(flow.udps.dst2src_flowpic_image)
        '''
        flow.udps.dst2src_flowpic_image[temp>0]  = 0
        flow.udps.dst2src_flowpic_image[temp==0] = 255
        flow.udps.dst2src_flowpic_image = flow.udps.dst2src_flowpic_image.astype(np.uint8)
        '''
        
        img = Image.fromarray(flow.udps.dst2src_flowpic_image , 'L') # L = 8bit black and white (gray scale)
        img.save(
            os.path.join(
                self.save_path,
                '-'.join(['flowpic',str(flow.bidirectional_first_seen_ms), flow.src_ip, str(flow.src_port), flow.dst_ip, str(flow.dst_port), str(flow.protocol), 'dst2src'])
                + '.png'))
        
        print(str(flow.bidirectional_first_seen_ms), np.sum(temp))
        
    def _scale_to(self, value_ms, min_value=0, max_value=1500):
        max_time_s = self.flow_active_time
        value_s = value_ms/1000
        return int(value_s*((max_value-min_value)/max_time_s))
    


import subprocess
import csv
import glob
import os
from datetime import datetime


def run_tshark(input_filepath, output_filepath, tshark_location, ip1, port1, ip2, port2, protocol, error_log_filepath='./tshark_error_log.txt', ):
    filter = f'{protocol} and ip.addr=={ip1} and {protocol}.port=={port1} and ip.addr=={ip2} and {protocol}.port=={port2}'
    splitted_save = f'-w {output_filepath}'.split()
    command =  f'{tshark_location} -r {input_filepath} -2 -n -R'
    splitted_command = command.split() + [filter] + splitted_save
    print(' '.join(splitted_command))
    with open(output_filepath, "w") as outfile:
        with open(error_log_filepath, "w") as error_log_file:
            proc = subprocess.run(
                splitted_command, 
                stdout = outfile, 
                stderr = error_log_file,
                check=True
            )

def read_flowpic_csv(filepath):
    session_tuple_keys = []
    with open(filepath, 'r') as csv_file:
        reader = csv.reader(csv_file)
        for i, row in enumerate(reader):
            session_tuple_key = [s.lower() for s in tuple(row[:8])]
            print(i, session_tuple_key)
            session_tuple_keys.append(session_tuple_key)
    return session_tuple_keys
            
def get_last_dir_in_path(dir_path):
    '''
    General Purpose
    '''
    return os.path.basename(os.path.normpath(dir_path))
             
def current_time():
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    return current_time



## main
filename_mapping = {
    'facebook_video1a': 'nonvpn-streaming-facebook-19.pcap',
    'facebook_video1b': 'nonvpn-streaming-facebook-20.pcap',
    'facebook_video2a': 'nonvpn-streaming-facebook-21.pcap',
    'facebook_video2b': 'nonvpn-streaming-facebook-22.pcap',
    'facebook_audio1a': 'nonvpn-voip-facebook-11.pcap',
    'facebook_audio1b': 'nonvpn-voip-facebook-12.pcap',
    'facebook_audio2a': 'nonvpn-voip-facebook-13.pcap',
    'facebook_audio2b': 'nonvpn-voip-facebook-14.pcap',
    'facebook_audio3':  'nonvpn-voip-facebook-15.pcap',
    'facebook_audio4':  'nonvpn-voip-facebook-16.pcap',
    'hangouts_audio1a': 'nonvpn-voip-hangouts-30.pcap',
    'hangouts_audio1b': 'nonvpn-voip-hangouts-31.pcap',
    'hangouts_audio2a': 'nonvpn-voip-hangouts-32.pcap',
    'hangouts_audio2b': 'nonvpn-voip-hangouts-33.pcap',
    'hangouts_audio3': 'nonvpn-voip-hangouts-34.pcap',
    'hangouts_audio4': 'nonvpn-voip-hangouts-35.pcap',
    'skype_audio1a': 'nonvpn-voip-skype-68.pcap',
    'skype_audio1b': 'nonvpn-voip-skype-69.pcap',
    'skype_audio2a': 'nonvpn-voip-skype-70.pcap',
    'skype_audio2b': 'nonvpn-voip-skype-71.pcap',
    'skype_audio3': 'nonvpn-voip-skype-72.pcap',
    'skype_audio4': 'nonvpn-voip-skype-73.pcap',
    'voipbuster1b': 'nonvpn-voip-voipbuster-96.pcap',
    'voipbuster2b': 'nonvpn-voip-voipbuster-97.pcap',
    'voipbuster3b': 'nonvpn-voip-voipbuster-98.pcap',
    'voipbuster_4a': 'nonvpn-voip-voipbuster-99.pcap',
    'voipbuster_4b': 'nonvpn-voip-voipbuster-100.pcap',
    'vpn_hangouts_audio1': '',
    'vpn_hangouts_audio1': '',
    'vpn_hangouts_audio2': '',
    'vpn_hangouts_audio2': '',
    'vpn_hangouts_audio2': '',
    'vpn_hangouts_audio2': '',
    'vpn_skype_audio1': '',
    'vpn_skype_audio1': '',
    'vpn_skype_audio2': '',
    'vpn_skype_audio2': '',
    'vpn_voipbuster1a': '',
    'vpn_voipbuster1a': '',
    'vpn_voipbuster1b': '',
    'vpn_voipbuster1b': '',
    'facebook_video1a': 'nonvpn-streaming-facebook-19.pcap',
    'facebook_video1b': 'nonvpn-streaming-facebook-20.pcap',
    'facebook_video2a': 'nonvpn-streaming-facebook-21.pcap',
    'facebook_video2b': 'nonvpn-streaming-facebook-22.pcap',
    'hangouts_video1b': 'nonvpn-streaming-hangouts-37.pcap',
    'hangouts_video2a': 'nonvpn-streaming-hangouts-38.pcap', # missing
    'hangouts_video2b': 'nonvpn-streaming-hangouts-39.pcap',
    'netflix1': 'nonvpn-streaming-netflix-44.pcap',
    'netflix2': 'nonvpn-streaming-netflix-45.pcap',
    'netflix3': 'nonvpn-streaming-netflix-46.pcap',
    'netflix4': 'nonvpn-streaming-netflix-47.pcap',
    'skype_video1a': 'nonvpn-streaming-skype-84.pcap', # one skype video is missing
    'skype_video1b': 'nonvpn-streaming-skype-85.pcap',
    'skype_video2a': 'nonvpn-streaming-skype-86.pcap',
    'skype_video2b': 'nonvpn-streaming-skype-87.pcap',
    'vimeo1': 'nonvpn-streaming-vimeo-92.pcap',
    'vimeo3': 'nonvpn-streaming-vimeo-94.pcap',
    'vimeo4': 'nonvpn-streaming-vimeo-95.pcap',
    'youtube5': 'nonvpn-streaming-youtube-135.pcap',
    'youtubeHTML5_1': '', # missing
    'ftps_down_1a': 'nonvpn-filetransfer-ftps-25.pcap',
    'ftps_down_1b': 'nonvpn-filetransfer-ftps-26.pcap',
    'scp1': 'nonvpn-filetransfer-scp- .pcap', # unknown
    'scpUp1': 'nonvpn-filetransfer-scp- .pcap',
    'scpUp2': 'nonvpn-filetransfer-scp- .pcap',
    'sftp_down_3a': 'nonvpn-filetransfer- - .pcap',
    'sftp_down_3b': 'nonvpn-filetransfer- - .pcap',
    'sftp_up_2a': 'nonvpn-filetransfer- - .pcap',
    'sftp_up_2b': 'nonvpn-filetransfer- - .pcap',
    'sftpUp1': 'nonvpn-filetransfer- - .pcap',
    'skype_file2': 'nonvpn-filetransfer-skype-76.pcap',
    'skype_file3': 'nonvpn-filetransfer-skype-77.pcap',
    'skype_file6': 'nonvpn-filetransfer-skype-78.pcap',
    'skype_file7': 'nonvpn-filetransfer-skype-81.pcap',
    'skype_file8': 'nonvpn-filetransfer-skype-82.pcap',


}

extracted_flows = set()

flowpic_csv_dir = r"flowpic_csvs/*.csv"
pcap_dir = r"/mnt/d/temp/University/masters/thesis/research/temp/FlowPic/Ofek's_implementation/data/nonvpn_cleaned"
output_dir = 'flowpic_extracted_flows'


print(current_time(), 'Starting')
for filepath in glob.glob(flowpic_csv_dir):
    output_subdir = os.path.join(output_dir, get_last_dir_in_path(filepath)[:-4])
    try:
        os.mkdir(output_subdir)
        print(current_time(), 'Directory created', output_subdir)
    except:
        print(current_time(), 'FAILED TO CREATE A DIRECTORY', output_subdir)
    keys = read_flowpic_csv(filepath)
    for key in keys:
        if tuple(sorted(key[0:6])) in extracted_flows:
            continue
        
        print(current_time(), 'Extracting with TShark', key)
        
        run_tshark(
            os.path.join(
                pcap_dir, 
                filename_mapping[key[0]]
            ),
            os.path.join(
                output_subdir, 
                '-'.join(key + ['.pcap'])
            ),
            'tshark',
            key[1],
            key[2],
            key[3],
            key[4],
            key[5]
        )
        
        extracted_flows.add(tuple(sorted(key[0:6])))
        


from stats.stats import IterableStats
import numpy as np
import pandas as pd
import hashlib 
import os
import json


class TLSRecordJoy():
    '''

    '''
    # Note that in Joy, the output path is always relative, i.e, it appends './' before the given filepath.
    def __init__(self, pcap_filepath, config_filepath=os.path.join(os.path.dirname(__file__), 'tools/config.json'), 
                 output_filepath='tls-cisco-joy.json', **kwargs):
        super().__init__(**kwargs)
        self.pcap_filepath = pcap_filepath
        self.config_filepath = config_filepath
        self.output_filepath = output_filepath

    def execute_joy(self, nfstream_sessions_df=None):
        joy_config = self._read_joy_config_file()
        joy_command = ' '.join([joy_config['CiscoJoyPath'], 
                                'preemptive_timeout=0', 
                                'bidir=1', 
                                'tls=1', 
                                'output='+self.output_filepath, 
                                self.pcap_filepath])
        print(joy_command)
        res = os.system(joy_command)
        if res != 0 :
            raise Exception('[3rd-party tool] Cisco Joy didnt execute succesfully')
        
        joy_df = self._read_joy_output()
        os.remove(self.output_filepath)
        if joy_df is None:
            return nfstream_sessions_df
        
        if nfstream_sessions_df is None:
            return joy_df
        
        nfstream_sessions_df['hashed_session_id'] = nfstream_sessions_df.apply(self._five_tuple, axis=1)
        nfstream_sessions_df.reset_index()
        
        merged_df = nfstream_sessions_df.merge(joy_df, 
                                               on='hashed_session_id',
                                               how='left',
                                               validate='one_to_one',
                                               left_index=True).drop('hashed_session_id', axis=1)
        
        return merged_df
        
    def _read_joy_config_file(self):
        with open(self.config_filepath) as text_config:
            json_config = json.loads(text_config.read())
            return json_config
    
    def _read_joy_output(self):
        exctracted_features = [] 
        with open(self.output_filepath, "r") as joy_output:
            for i,row in enumerate(joy_output):
                if i == 0: # first line is metadata
                    continue # skip
                flow = json.loads(row)
                if 'tls' in flow:
                    exctracted_features.append(self._extract_features(flow))
                
        if len(exctracted_features) == 0:
            return None
        
        joy_df = pd.DataFrame(exctracted_features)
        joy_df['protocol'] = 6 # TCP
        joy_df['hashed_session_id'] = joy_df.apply(self._five_tuple, axis=1)
        joy_df = joy_df.drop(['src_ip', 'src_port','dst_ip','dst_port','protocol'], axis=1) 
        
        return joy_df.reset_index()
    
    def _extract_features(self, flow, n_records=20):
        # general flow features
        extracted_flow_features = {
            'src_ip':   flow['sa'], 
            'src_port': flow['sp'], 
            'dst_ip':   flow['da'], 
            'dst_port': flow['dp'],
        }
        
        # TLS:
        tls = flow['tls']
        exctracted_general_tls_features = self._exctract_general_tls_features(tls)
        exctracted_tls_features = self._exctract_tls_record_features(tls['srlt'])
        extracted_early_tls_features = self._exctract_tls_record_features(tls['srlt'][:min(n_records, len(tls['srlt']))])
        # rename dictionary keys
        extracted_early_tls_features = { key+'_n' : value for key, value in extracted_early_tls_features.items() }
        
        return {**exctracted_general_tls_features,
                **extracted_flow_features,
                **exctracted_tls_features,
                'tls_early_records_n': n_records,
                **extracted_early_tls_features} 
    
    def _exctract_general_tls_features(self, tls):
        # general TLS features
        extracted_tls_general_features = {
            'tls_cipher_suites': list(tls['cs']),
        }
        
        return extracted_tls_general_features
    
    def _extract_tls_clump_features(self, clumps):
        # clumped TLS features
        '''
        Clump is an aggregation of multiple records in the same direction.
        A clump is closed when the following record is sent in the reversed direction, which then
        opens a new clump for records in the reversed direction.
        
        The first clump is from source to destination.
        '''
        # bidirectional
        clump_bytes = list(map(lambda clump: sum(map(lambda record: record['b'], clump)), clumps))
        clump_bytes_stats = IterableStats(clump_bytes)
        clump_sizes = list(map(lambda clump: len(clump), clumps))
        clump_sizes_stats = IterableStats(clump_sizes)
        extracted_tls_clumped_features = {
            'bidirectional_tls_clumps': len(clumps),
            'bidirectional_tls_clump_bytes': clump_bytes,
            'bidirectional_tls_clump_sizes': clump_sizes,
            'bidirectional_mean_tls_clump_bytes':             clump_bytes_stats.average(),
            'bidirectional_median_tls_clump_bytes':           clump_bytes_stats.median(),
            'bidirectional_stddev_tls_clump_bytes':           clump_bytes_stats.std_deviation(),
            'bidirectional_variance_tls_clump_bytes':         clump_bytes_stats.variance(),
            'bidirectional_skew_from_median_tls_clump_bytes': clump_bytes_stats.skew_from_median(),
            'bidirectional_coeff_of_var_tls_clump_bytes':     clump_bytes_stats.coeff_of_variation(),
            'bidirectional_min_tls_clump_bytes':              clump_bytes_stats.min(),
            'bidirectional_max_tls_clump_bytes':              clump_bytes_stats.max(),
            'bidirectional_mean_tls_clump_sizes':             clump_sizes_stats.average(),
            'bidirectional_median_tls_clump_sizes':           clump_sizes_stats.median(),
            'bidirectional_stddev_tls_clump_sizes':           clump_sizes_stats.std_deviation(),
            'bidirectional_variance_tls_clump_sizes':         clump_sizes_stats.variance(),
            'bidirectional_skew_from_median_tls_clump_sizes': clump_sizes_stats.skew_from_median(),
            'bidirectional_coeff_of_var_tls_clump_sizes':     clump_sizes_stats.coeff_of_variation(),
            'bidirectional_min_tls_clump_sizes':              clump_sizes_stats.min(),
            'bidirectional_max_tls_clump_sizes':              clump_sizes_stats.max(),
        }
        # src -> dst
        src2dst_clumps = clumps[::2]
        src2dst_clump_bytes = list(map(lambda clump: sum(map(lambda record: record['b'], clump)), src2dst_clumps))
        src2dst_clump_bytes_stats = IterableStats(src2dst_clump_bytes)
        src2dst_clump_sizes = list(map(lambda clump: len(clump), src2dst_clumps))
        src2dst_clump_sizes_stats = IterableStats(src2dst_clump_sizes)
        src2dst_extracted_tls_clumped_features = {
            'src2dst_tls_clumps': len(src2dst_clumps),
            'src2dst_tls_clump_bytes': src2dst_clump_bytes,
            'src2dst_tls_clump_sizes': src2dst_clump_sizes,
            'src2dst_mean_tls_clump_bytes':             src2dst_clump_bytes_stats.average(),
            'src2dst_median_tls_clump_bytes':           src2dst_clump_bytes_stats.median(),
            'src2dst_stddev_tls_clump_bytes':            src2dst_clump_bytes_stats.std_deviation(),
            'src2dst_variance_tls_clump_bytes':         src2dst_clump_bytes_stats.variance(),
            'src2dst_skew_from_median_tls_clump_bytes': src2dst_clump_bytes_stats.skew_from_median(),
            'src2dst_coeff_of_var_tls_clump_bytes':     src2dst_clump_bytes_stats.coeff_of_variation(),
            'src2dst_min_tls_clump_bytes':              src2dst_clump_bytes_stats.min(),
            'src2dst_max_tls_clump_bytes':              src2dst_clump_bytes_stats.max(),
            'src2dst_mean_tls_clump_sizes':             src2dst_clump_sizes_stats.average(),
            'src2dst_median_tls_clump_sizes':           src2dst_clump_sizes_stats.median(),
            'src2dst_stddev_tls_clump_sizes':            src2dst_clump_sizes_stats.std_deviation(),
            'src2dst_variance_tls_clump_sizes':         src2dst_clump_sizes_stats.variance(),
            'src2dst_skew_from_median_tls_clump_sizes': src2dst_clump_sizes_stats.skew_from_median(),
            'src2dst_coeff_of_var_tls_clump_sizes':     src2dst_clump_sizes_stats.coeff_of_variation(),
            'src2dst_min_tls_clump_sizes':              src2dst_clump_sizes_stats.min(),
            'src2dst_max_tls_clump_sizes':              src2dst_clump_sizes_stats.max(),
        }
        # dst -> src
        dst2src_clumps = clumps[1::2]
        dst2src_clump_bytes = list(map(lambda clump: sum(map(lambda record: record['b'], clump)), dst2src_clumps))
        dst2src_clump_bytes_stats = IterableStats(dst2src_clump_bytes)
        dst2src_clump_sizes = list(map(lambda clump: len(clump), dst2src_clumps))
        dst2src_clump_sizes_stats = IterableStats(dst2src_clump_sizes)
        dst2src_extracted_tls_clumped_features = {
            'dst2src_tls_clumps': len(dst2src_clumps),
            'dst2src_tls_clump_bytes': dst2src_clump_bytes,
            'dst2src_tls_clump_sizes': dst2src_clump_sizes,
            'dst2src_mean_tls_clump_bytes':             dst2src_clump_bytes_stats.average(),
            'dst2src_median_tls_clump_bytes':           dst2src_clump_bytes_stats.median(),
            'dst2src_stddev_tls_clump_bytes':            dst2src_clump_bytes_stats.std_deviation(),
            'dst2src_variance_tls_clump_bytes':         dst2src_clump_bytes_stats.variance(),
            'dst2src_skew_from_median_tls_clump_bytes': dst2src_clump_bytes_stats.skew_from_median(),
            'dst2src_coeff_of_var_tls_clump_bytes':     dst2src_clump_bytes_stats.coeff_of_variation(),
            'dst2src_min_tls_clump_bytes':              dst2src_clump_bytes_stats.min(),
            'dst2src_max_tls_clump_bytes':              dst2src_clump_bytes_stats.max(),
            'dst2src_mean_tls_clump_sizes':             dst2src_clump_sizes_stats.average(),
            'dst2src_median_tls_clump_sizes':           dst2src_clump_sizes_stats.median(),
            'dst2src_stddev_tls_clump_sizes':            dst2src_clump_sizes_stats.std_deviation(),
            'dst2src_variance_tls_clump_sizes':         dst2src_clump_sizes_stats.variance(),
            'dst2src_skew_from_median_tls_clump_sizes': dst2src_clump_sizes_stats.skew_from_median(),
            'dst2src_coeff_of_var_tls_clump_sizes':     dst2src_clump_sizes_stats.coeff_of_variation(),
            'dst2src_min_tls_clump_sizes':              dst2src_clump_sizes_stats.min(),
            'dst2src_max_tls_clump_sizes':              dst2src_clump_sizes_stats.max(),
        }
        
        return {**src2dst_extracted_tls_clumped_features,
                **dst2src_extracted_tls_clumped_features,
                **extracted_tls_clumped_features}
        
    
    def _exctract_tls_record_features(self, records):
        # src -> dst
        src2dst_tls_records = list(filter(lambda tls_record: tls_record['dir'] == '>', records))
        src2dst_tls_record_sizes = list(map(lambda tls_record: tls_record['b'], src2dst_tls_records))
        src2dst_stats = IterableStats(src2dst_tls_record_sizes)
        src2dst_extracted_features = {
            'src2dst_tls_records': len(src2dst_tls_records),
            'src2dst_tls_record_types': list(map(lambda tls_record: tls_record['tp'], src2dst_tls_records)) ,
            'src2dst_tls_record_sizes': src2dst_tls_record_sizes,
            'src2dst_tls_record_frequencies': {key : value for key,value in np.asarray(np.unique(src2dst_tls_record_sizes, return_counts=True) ).T},
            'src2dst_tls_payload_bytes': sum(src2dst_tls_record_sizes),
            'src2dst_tls_record_distinct_sizes': len(np.unique(src2dst_tls_record_sizes)),
            'src2dst_mean_tls_record_size':             src2dst_stats.average(),
            'src2dst_median_tls_record_size':           src2dst_stats.median(),
            'src2dst_stddev_tls_record_size':           src2dst_stats.std_deviation(),
            'src2dst_variance_tls_record_size':         src2dst_stats.variance(),
            'src2dst_skew_from_median_tls_record_size': src2dst_stats.skew_from_median(),
            'src2dst_coeff_of_var_tls_record_size':     src2dst_stats.coeff_of_variation(),
            'src2dst_min_tls_record_size':              src2dst_stats.min(),
            'src2dst_max_tls_record_size':              src2dst_stats.max(),
        }
        
        # dst -> src
        dst2src_tls_records = list(filter(lambda tls_record: tls_record['dir'] == '<', records))                                                  
        dst2src_tls_record_sizes = list(map(lambda tls_record: tls_record['b'], dst2src_tls_records))
        dst2src_stats = IterableStats(dst2src_tls_record_sizes)
        dst2src_extracted_features = {
            'dst2src_tls_records': len(dst2src_tls_records),
            'dst2src_tls_record_types': list(map(lambda tls_record: tls_record['tp'], dst2src_tls_records)) ,
            'dst2src_tls_record_sizes': dst2src_tls_record_sizes,
            'dst2src_tls_record_frequencies': {key : value for key,value in np.asarray(np.unique(dst2src_tls_record_sizes, return_counts=True) ).T},
            'dst2src_tls_payload_bytes': sum(dst2src_tls_record_sizes),
            'dst2src_tls_record_distinct_sizes': len(np.unique(dst2src_tls_record_sizes)),
            'dst2src_mean_tls_record_size':             dst2src_stats.average(),
            'dst2src_median_tls_record_size':           dst2src_stats.median(),
            'dst2src_stddev_tls_record_size':            dst2src_stats.std_deviation(),
            'dst2src_variance_tls_record_size':         dst2src_stats.variance(),
            'dst2src_skew_from_median_tls_record_size': dst2src_stats.skew_from_median(),
            'dst2src_coeff_of_var_tls_record_size':     dst2src_stats.coeff_of_variation(),
            'dst2src_min_tls_record_size':              dst2src_stats.min(),
            'dst2src_max_tls_record_size':              dst2src_stats.max(),
            
        }
        # bidirectional                                    
        tls_record_sizes = list(map(lambda tls_record: tls_record['b'],  records))
        bi_stats = IterableStats(tls_record_sizes)
        bidirectional_extracted_features = {
            'bidirectional_tls_records': len(records),
            'bidirectional_tls_record_types': list(map(lambda tls_record: tls_record['tp'], records)) ,
            'bidirectional_tls_record_sizes': tls_record_sizes,
            'bidirectional_tls_record_frequencies': {key : value for key,value in np.asarray(np.unique(tls_record_sizes, return_counts=True) ).T},
            'bidirectional_tls_payload_bytes': sum(tls_record_sizes),
            'bidirectional_tls_record_distinct_sizes': len(np.unique(tls_record_sizes)),
            'bidirectional_mean_tls_record_size':             bi_stats.average(),
            'bidirectional_median_tls_record_size':           bi_stats.median(),
            'bidirectional_stddev_tls_record_size':           bi_stats.std_deviation(),
            'bidirectional_variance_tls_record_size':         bi_stats.variance(),
            'bidirectional_skew_from_median_tls_record_size': bi_stats.skew_from_median(),
            'bidirectional_coeff_of_var_tls_record_size':     bi_stats.coeff_of_variation(),
            'bidirectional_min_tls_record_size':              bi_stats.min(),
            'bidirectional_max_tls_record_size':              bi_stats.max(),
        }
        
        return {**src2dst_extracted_features, 
                **dst2src_extracted_features,
                **bidirectional_extracted_features,
                **self._extract_tls_clump_features(self._clumps(records))}
        
    def _clumps(self, tls_records):
        clumps = []
        dir = '>'
        begin = 0
        end = 0
        for i, record in enumerate(tls_records):
            if dir != record['dir']:
                end = i
                clump = tls_records[begin:end]
                clumps.append(clump)
                begin = end
                dir = record['dir']
        
        if end < begin:
            clump = tls_records[begin:len(tls_records)]
            clumps.append(clump)
            
        return clumps

    def _five_tuple(self, row):
        return hashlib.md5(
            '-'.join(sorted([
                row['src_ip'],
                str(row['src_port']),
                row['dst_ip'],
                str(row['dst_port']),
                str(row['protocol'])
            ])).encode()
        ).digest()
        
        
        
        
if __name__ == '__main__':
    print('starting')
    #df = pd.read_csv('./temp/out-sessions.csv')
    merged_df = TLSRecordJoy(r'./pcaps/DoH-Firefox84-NextDNS-1-pcap-format.pcap').execute_joy()
    
    merged_df.to_csv('./temp/out-sessions-merged-with-tls.csv')




# from tls_record_joy import TLSRecordJoy


from tls_tshark_entry import extract_tls_features_and_merge
from hosts_processor import HostsProcessor
from .plugins.clump_flows import Clump_Flow
from .plugins.packets_size_interarrival_time import Packets_size_and_interarrival_time
from .plugins.asn_info import ASNInfo
from .plugins.n_pkts_byte_freq import NPacketsByteFrequency
from .plugins.first_packet_payload import FirstPacketPayloadLen
from .plugins.most_freq_payload_len_ratio import MostFreqPayloadLenRatio
from .plugins.dns_counter import DNSCounter
from .plugins.small_pkt_payload_ratio import SmallPacketPayloadRatio
from .plugins.pkt_rel_time import PacketRelativeTime
from .plugins.res_req_diff_time import ResReqDiffTime
from .plugins.graypic import GrayPic1
from .plugins.protocol_header_fields import ProtocolHeaderFields
from .plugins.n_bytes import NBytes
from .plugins.stnn import STNN
from nfstream import NFStreamer  # https://www.nfstream.org/docs/api
from os import path
#from guppy import hpy; h=hpy()


class Extractor:
    def __init__(self, input_pcap_filepath, output_dirpath, bpf_filter_string=None, custom_plugin_package=None, TLS=True) -> None:
        self.input_pcap_filepath = input_pcap_filepath
        self.output_dirpath = output_dirpath
        self.bpf_filter_string = bpf_filter_string
        self.custom_plugin_package = custom_plugin_package
        self.TLS = TLS
        
    def extract(self):
        plugins = self._plugins() if self.custom_plugin_package is None else self.custom_plugin_package

        # Time-windowed flows
        '''
        my_streamer = NFStreamer(source=self.input_pcap_filepath,
                                decode_tunnels=True,
                                bpf_filter=self.bpf_filter_string,
                                promiscuous_mode=True,
                                snapshot_length=1536,
                                idle_timeout=1,
                                active_timeout=1,
                                accounting_mode=3,
                                udps=plugins,
                                n_dissections=20,
                                statistical_analysis=True,
                                splt_analysis=0,
                                n_meters=0,
                                performance_report=0)
        
        my_streamer.to_csv(path.join(self.output_dirpath,'out-timed-flows.csv'), columns_to_anonymize=[])
        del my_streamer
        '''
        
        # Sessions
        session_streamer = NFStreamer(source=self.input_pcap_filepath,
                                decode_tunnels=True,
                                bpf_filter=self.bpf_filter_string,
                                promiscuous_mode=True,
                                snapshot_length=1536,
                                idle_timeout=999999999,
                                active_timeout=999999999,
                                accounting_mode=0,
                                udps=plugins,
                                n_dissections=20,
                                statistical_analysis=True,
                                splt_analysis=0,
                                n_meters=0,
                                performance_report=0)

        if self.TLS:
            sessions_df = session_streamer.to_pandas(columns_to_anonymize=[])
            config_filepath = path.join(path.dirname(__file__), 'tools/config.json')
            sessions_df = extract_tls_features_and_merge(sessions_df, self.input_pcap_filepath, config_filepath)
            sessions_df.to_csv(path.join(self.output_dirpath,'out-sessions.csv'), index=False)
        else:
            session_streamer.to_csv(path.join(self.output_dirpath,'out-sessions.csv'))

        # Hosts
        # HostsProcessor(sessions_df, self.output_dirpath).process()
        #print(6,h.heap())
        return path.join(self.output_dirpath,'out-sessions.csv')

    def _plugins(self):
        return [
            ASNInfo(),
            DNSCounter(),
            FirstPacketPayloadLen(),
            MostFreqPayloadLenRatio(),
            NPacketsByteFrequency(n_first_packets=6),
            PacketRelativeTime(),
            SmallPacketPayloadRatio(),
            ResReqDiffTime(),
            Clump_Flow(),
            Packets_size_and_interarrival_time(),
            ProtocolHeaderFields(20),
            NBytes(),
            STNN(20)
        ]


from nfstream import NFPlugin

class FirstPacketPayloadLen(NFPlugin):
    '''
        First packet's payload length per direction,
        note that the first packet is always from src to dst hence
        bidirectional first packet's payload length is equal to src2dst_packet_payload_len.
        
        if there are no packets from dst to src, then the value is defaulted to None (empty).
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        '''
        on_init(self, packet, flow): Method called at flow creation.
        '''
        flow.udps.src2dst_first_packet_payload_len = packet.payload_size
        flow.udps.dst2src_first_packet_payload_len = None
        
    def on_update(self, packet, flow):
        if flow.udps.dst2src_first_packet_payload_len is None and packet.direction == 1:
           flow.udps.dst2src_first_packet_payload_len = packet.payload_size 



class SessionsProcessor:
    def __init__(self) -> None:
        pass
    
    def process(self, df):
        grouped_by_session = df.groupby(lambda f: self._five_tuple(df.iloc[f]))
        
        
        print('Finished processing sessions')


    def _five_tuple(self, flow):
        return '-'.join(
            sorted([
                flow.src_ip,
                str(flow.src_port),
                flow.dst_ip,
                str(flow.dst_port),
                str(flow.protocol)
            ])
        )
        
    def _fill_empty_timed_windows(self, time_windowed_flows):
        '''
        W.I.P
        '''
        previous_window = time_windowed_flows[0]
        for i in range(1, len(time_windowed_flows)):
            current_window = time_windowed_flows[1]
            time_diff = current_window.bidirectional_first_seen_ms - previous_window.bidirectional_last_seen_ms + (1 - previous_window.bidirectional_duration_ms)
    


from third_party.tshark.dataframe_utils import add_packet_clump_num_column, add_packet_directions_column, convert_to_correct_dtypes, drop_lengthless_tls_records, get_tls_clump_stats, get_tls_record_stats, groupby_biflows, merge_df_by_biflows, read_tshark_csv_output
tshark_output_filepath = './temp/tshark_outfile.csv'
df = read_tshark_csv_output(tshark_output_filepath)
df = drop_lengthless_tls_records(df)
if df.empty:
    tls_features_per_session = None
else:   
    df = convert_to_correct_dtypes(df)
    df = add_packet_directions_column(df)
    df = add_packet_clump_num_column(df)
    record_tls_per_session = get_tls_record_stats(df)
    clump_stats_per_session = get_tls_clump_stats(df)
    tls_features_per_session = record_tls_per_session.join(clump_stats_per_session)


from nfstream import NFStreamer, NFPlugin
import numpy as np
from scapy.all import IP, IPv6, raw

class NBytes(NFPlugin):
    '''
        Extracts the first n_bytes from the flow, the bytes are taken
        from the transport layer payload (L4). if the flow have less than n_bytes bytes,
        then the rest of the bytes are zero-valued (padding).
    '''
    def __init__(self, n=784):
        self.n = n
    
    def on_init(self, packet, flow):
        flow.udps.n_bytes_value = self.n
        flow.udps.n_bytes = [] # np.zeros(self.n)
        flow.udps.n_bytes_counted = 0
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        remaining_bytes = self.n - flow.udps.n_bytes_counted
        if remaining_bytes >= 0 and packet.protocol in [6, 17]: # TCP or UDP only.
            amount_to_copy = min(remaining_bytes, packet.payload_size)
            if amount_to_copy == 0:
                return
           
            flow.udps.n_bytes.extend(
                self.get_payload_as_binary_scapy(packet, packet.ip_version)[:amount_to_copy]
            )
            flow.udps.n_bytes_counted += amount_to_copy

    def on_expire(self, flow):
        '''
        Normalize to [0,1]: 
        flow.udps.n_bytes /= 255
        Optional cleanup: 
        del flow.udps.n_bytes_counted
        '''
        # Padding if necessary.
        if flow.udps.n_bytes_counted < self.n:
            remaining_bytes = self.n - flow.udps.n_bytes_counted
            flow.udps.n_bytes.extend(
                np.full(remaining_bytes, 0)
            )

    def get_payload_as_binary(self, packet):
        return packet.ip_packet[-packet.payload_size:]
    
    def get_payload_as_binary_scapy(self, packet, ip_version):
        '''
        Older versions of NFStream (lower than 6.1) had problems with 
        detection the correct payload size of ipv6 packets.
        This function is here as a fallback in case of problems in future
        versions.

        in 6.4.2 and 6.4.3 there is a bug with wrong transport_size, also
        NFStream does not detect ethernet padding.
        '''
        if ip_version == 4:
            scapy_packet = IP(packet.ip_packet)
        elif ip_version == 6:
            scapy_packet = IPv6(packet.ip_packet)

        return raw(scapy_packet.payload.payload)
        
